import os
import pytest
import jsonschema
import copy
import numpy as np
import numpy.testing as npt
import time as ttime
import re
from pyxrf.core.utils import convert_time_from_nexus_string
from pyxrf.core.xrf_utils import validate_element_str, generate_eline_list, split_compound_mass
from pyxrf.core.quant_analysis import (
    save_xrf_standard_yaml_file,
    load_xrf_standard_yaml_file,
    _xrf_standard_schema,
    load_included_xrf_standard_yaml_file,
    compute_standard_element_densities,
    _xrf_quant_fluor_schema,
    save_xrf_quant_fluor_json_file,
    load_xrf_quant_fluor_json_file,
    get_quant_fluor_data_dict,
    fill_quant_fluor_data_dict,
    prune_quant_fluor_data_dict,
    set_quant_fluor_data_dict_optional,
    set_quant_fluor_data_dict_time,
    ParamQuantEstimation,
    ParamQuantitativeAnalysis,
)

# Short example of XRF standard data
_standard_data_sample = [
    {
        "name": "Test Micromatter 411570",
        "serial": "411570",
        "description": "InSx 22.4 (In=16.0 S=6.4) / SrF2 20.6 / Cr 19.8 / Ni 19.2 / GaAs 19.1 (Ga=8.3 As=10.8)",
        "compounds": {"In": 16.0, "S": 6.4, "SrF2": 20.6, "Cr": 19.8, "Ni": 19.2, "Ga": 8.3, "As": 10.8},
        "density": 101.1,
    },
    {
        "name": "Test Micromatter 411640",
        "serial": "411640",
        "description": "CeF3 21.1 / Au 20.6",
        "compounds": {"CeF3": 21.1, "Au": 20.6}
        # Missing optional 'density' field
    },
]


# fmt: off
@pytest.mark.parametrize("standard_data", [
    _standard_data_sample,
    []  # The function should also work if the list is empty
])
# fmt: on
def test_save_xrf_standard_yaml_file1(tmp_path, standard_data):
    r"""Basic test of function 'save_xrf_standard_yaml_file' and 'load_xrf_standard_yaml_file'"""

    yaml_path = ["yaml", "param", "file"]
    file_name = "standard.yaml"
    yaml_path = os.path.join(tmp_path, *yaml_path, file_name)

    # Sample data
    save_xrf_standard_yaml_file(yaml_path, standard_data)

    data_loaded = load_xrf_standard_yaml_file(yaml_path)

    assert data_loaded == standard_data, "Loaded data is not equal to the original data"


def test_save_xrf_standard_yaml_file2(tmp_path):
    r"""Test the case of of no match between total density and sum of densities of individual compounds"""

    yaml_path = ["yaml", "param", "file"]
    file_name = "standard.yaml"
    yaml_path = os.path.join(tmp_path, *yaml_path, file_name)

    standard_data = []  # No records, but the file will still be saved

    # Sample data
    save_xrf_standard_yaml_file(yaml_path, standard_data)

    # Try to overwrite the file ('re.escape' is required when tests are run on Windows)
    with pytest.raises(IOError, match=f"File '{re.escape(yaml_path)}' already exists"):
        save_xrf_standard_yaml_file(yaml_path, standard_data)

    # The following should work
    save_xrf_standard_yaml_file(yaml_path, standard_data, overwrite_existing=True)


def test_save_xrf_standard_yaml_file3(tmp_path):
    r"""Test the case of existing file"""

    yaml_path = ["yaml", "param", "file"]
    file_name = "standard.yaml"
    yaml_path = os.path.join(tmp_path, *yaml_path, file_name)

    standard_data = copy.deepcopy(_standard_data_sample)
    standard_data[0]["density"] = 50.0  # Wrong value of total density

    # Save incorrect data
    save_xrf_standard_yaml_file(yaml_path, standard_data)
    #    and now try to read it
    with pytest.raises(RuntimeError, match="Sum of areal densities does not match total density"):
        load_xrf_standard_yaml_file(yaml_path)


def test_load_xrf_standard_yaml_file1(tmp_path):
    r"""Try loading non-existent YAML file"""

    file_name = "standard.yaml"

    # Try loading from the existing directory
    # 're.escape' is necessary when test are run on Windows
    yaml_path = os.path.join(tmp_path, file_name)
    with pytest.raises(IOError, match=f"File '{re.escape(yaml_path)}' does not exist"):
        load_xrf_standard_yaml_file(yaml_path)

    # Try loading from the non-existing directory
    yaml_path = ["yaml", "param", "file"]
    yaml_path = os.path.join(tmp_path, *yaml_path, file_name)
    with pytest.raises(IOError, match=f"File '{re.escape(yaml_path)}' does not exist"):
        load_xrf_standard_yaml_file(yaml_path)


def test_load_xrf_standard_yaml_file2(tmp_path):
    r"""Test for reporting schema violation"""

    yaml_path = ["yaml", "param", "file"]
    file_name = "standard.yaml"
    yaml_path = os.path.join(tmp_path, *yaml_path, file_name)

    standard_data = copy.deepcopy(_standard_data_sample)
    # Change name from string to number (this will not satisfy the built-in schema)
    for v in standard_data:
        v["name"] = 50.36

    # Save the changed dataset
    save_xrf_standard_yaml_file(yaml_path, standard_data)

    with pytest.raises(jsonschema.ValidationError):
        load_xrf_standard_yaml_file(yaml_path)

    # Disable schema validation by setting 'schema=None'
    load_xrf_standard_yaml_file(yaml_path, schema=None)

    # Change the schema to match data and validate with changed schema
    schema = copy.deepcopy(_xrf_standard_schema)
    schema["properties"]["name"] = {"type": ["string", "number"]}
    standard_data_out = load_xrf_standard_yaml_file(yaml_path, schema=schema)
    assert standard_data_out == standard_data, "Loaded data is different from the original"


def test_load_included_xrf_standard_yaml_file():

    data = load_included_xrf_standard_yaml_file()
    assert len(data) > 1, "Standard YAML file can not be read"


def test_compute_standard_element_densities():

    standard_data = _standard_data_sample

    for data in standard_data:

        # Find total density
        if "density" in data:
            total_density = data["density"]
        else:
            total_density = np.sum(list(data["compounds"].values()))

        element_densities = compute_standard_element_densities(data["compounds"])

        # Validate that all the keys of the returned dictionary represent elements
        assert all(
            [validate_element_str(_) for _ in element_densities.keys()]
        ), f"Some of the elements in the list {list(element_densities.keys())} have invalid representation"

        # Check if the sum of all return densities equals total density
        npt.assert_almost_equal(
            np.sum(list(element_densities.values())),
            total_density,
            err_msg="The sum of element densities and the total sum don't match",
        )


# -----------------------------------------------------------------------------------------------------


# Short example of XRF standard quantitative data record
#   The data has no physical meaning, used for testing of saving/loading of JSON file
_xrf_standard_fluor_sample = {
    "name": "Hypothetical sample #41157",
    "serial": "41157",
    "description": "Name of hypothetical sample",
    "element_lines": {
        "In": {"density": 16.0, "fluorescence": 1.5453452},
        "S_K": {"density": 6.4, "fluorescence": 2.0344345},
        "Sr_L": {"density": 20.6, "fluorescence": 0.93452365},
        "Au_M": {"density": 10.4, "fluorescence": 0.734234},
        "Cr_Ka": {"density": 19.8, "fluorescence": 0.7435234},
        "Ni_Kb": {"density": 19.2, "fluorescence": 0.7435234},
        "Ga_Ka1": {"density": 8.3, "fluorescence": 0.7435234},
        "Mg_Ka2": {"density": 9.6, "fluorescence": None},
    },
    "incident_energy": 10.5,
    "detector_channel": "sum",
    "scaler_name": "i0",
    "distance_to_sample": 50.6,
    "creation_time_local": "2020-01-10T11:50:39+00:00",
    "source_scan_id": 92276,
    "source_scan_uid": "f07e3065-ab92-4b20-a702-ef61ed164dbc",
}


def _get_data_and_json_path(tmp_path):

    # Create some complicated path
    json_path = ["json", "param", "file"]
    file_name = "standard.yaml"
    json_path = os.path.join(tmp_path, *json_path, file_name)

    data = _xrf_standard_fluor_sample

    return data, json_path


def test_save_xrf_quant_fluor_json_file1(tmp_path):
    r"""Basic test of function 'save_xrf_standard_yaml_file' and 'load_xrf_standard_yaml_file'"""

    data, json_path = _get_data_and_json_path(tmp_path)

    # Sample data
    save_xrf_quant_fluor_json_file(json_path, data)

    data_loaded = load_xrf_quant_fluor_json_file(json_path)

    assert data_loaded == data, "Loaded data is not equal to the original data"


def test_save_xrf_quant_fluor_json_file2(tmp_path):
    # 'save_xrf_quant_fluor_json_file' - overwrite existing file

    data, json_path = _get_data_and_json_path(tmp_path)

    # Create file
    save_xrf_quant_fluor_json_file(json_path, data)

    # Attempt to overwrite (note: 're.escape' is necessary for Windows paths)
    with pytest.raises(IOError, match=f"File '{re.escape(json_path)}' already exists"):
        save_xrf_quant_fluor_json_file(json_path, data)

    # Now overwrite the file by setting the flag 'overwrite_existing=True'
    save_xrf_quant_fluor_json_file(json_path, data, overwrite_existing=True)


def test_save_xrf_quant_fluor_json_file3(tmp_path):
    # 'save_xrf_quant_fluor_json_file' - invalid schema

    data, json_path = _get_data_and_json_path(tmp_path)

    data = copy.deepcopy(data)
    # Change the data so that it doesn't satisfy the schema
    data["incident_energy"] = "incident_energy"  # Supposed to be a number

    with pytest.raises(jsonschema.ValidationError):
        save_xrf_quant_fluor_json_file(json_path, data)


def test_save_xrf_quant_fluor_json_file4(tmp_path):
    r"""Schema allows some data fields to hold value of ``None``. Test if this works."""
    data, json_path = _get_data_and_json_path(tmp_path)

    # Modify some elements of the dictionary
    data = copy.deepcopy(data)
    data["detector_channel"] = None
    data["scaler_name"] = None
    data["distance_to_sample"] = None

    # Sample data
    save_xrf_quant_fluor_json_file(json_path, data)

    data_loaded = load_xrf_quant_fluor_json_file(json_path)

    assert data_loaded == data, "Loaded data is not equal to the original data"


def test_load_xrf_quant_fluor_json_file1(tmp_path):
    # 'load_xrf_quant_fluor_json_file' - non-existing file

    _, json_path = _get_data_and_json_path(tmp_path)

    # 're.escape' is necessary if test is run on Windows
    with pytest.raises(IOError, match=f"File '{re.escape(json_path)}' does not exist"):
        load_xrf_quant_fluor_json_file(json_path)


def test_load_xrf_quant_fluor_json_file2(tmp_path):
    # 'load_xrf_quant_fluor_json_file' - schema is not matched

    data, json_path = _get_data_and_json_path(tmp_path)

    # Create file
    save_xrf_quant_fluor_json_file(json_path, data)

    schema = copy.deepcopy(_xrf_quant_fluor_schema)
    # Modify schema, so that it is incorrect
    schema["properties"]["scaler_name"] = {"type": "number"}  # Supposed to be a string

    with pytest.raises(jsonschema.ValidationError):
        load_xrf_quant_fluor_json_file(json_path, schema=schema)


def test_get_quant_fluor_data_dict():
    """Tests for 'get_quant_fluor_data_dict': basic tests for consistensy of the returned dictionary"""

    for standard_data in _standard_data_sample:

        quant_fluor_data_dict = get_quant_fluor_data_dict(standard_data, incident_energy=12.0)
        # Will raise exception is schema is not satisfied
        jsonschema.validate(instance=quant_fluor_data_dict, schema=_xrf_quant_fluor_schema)

        assert quant_fluor_data_dict["name"] == standard_data["name"], "Dictionary element 'name' is incorrect"

        assert (
            quant_fluor_data_dict["serial"] == standard_data["serial"]
        ), "Dictionary element 'serial' is incorrect"

        assert (
            quant_fluor_data_dict["description"] == standard_data["description"]
        ), "Dictionary element 'description' is incorrect"

        eline_set = set()
        # The 'mass' is not actual mass. If elements has multiple emission lines activated, then
        #   the mass (density) of the element will be counted multiple times. There is no
        #   physical meaning in the computed value: it is used to verify the sum of densities in
        #   the 'element_lines' dictionary of emission lines in 'quant_fluor_data_dict'
        mass_sum_expected = 0
        for cmpd, mass in standard_data["compounds"].items():
            em_dict = split_compound_mass(cmpd, mass)
            for el, ms in em_dict.items():
                elines = generate_eline_list([el], incident_energy=12.0)
                n_elines = len(elines)
                if n_elines:
                    mass_sum_expected += n_elines * ms
                    eline_set.update(elines)

        eline_out_list = list(quant_fluor_data_dict["element_lines"].keys())
        assert len(eline_out_list) == len(eline_set), "The number of emission lines is not as expected"
        assert (
            set(eline_out_list) == eline_set
        ), "Generated object contains emission lines that are different from expected"

        mass_sum = sum([_["density"] for _ in quant_fluor_data_dict["element_lines"].values()])
        assert (
            mass_sum == mass_sum_expected
        ), "The total mass (density) of the components is different from expected"


def gen_xrf_map_dict(nx=10, ny=5, elines=["S_K", "Au_M", "Fe_K"]):
    r"""
    Create dictionary with the given set of emission lines and scaler ``sclr``
    """
    img = {}
    for e in elines:
        map = np.random.rand(ny, nx) * np.random.rand() * 10
        img[e] = map

    # Scaler
    map_sclr = np.ones(shape=(ny, nx), dtype=float) * np.random.rand() * 2
    img["sclr"] = map_sclr

    return img


# fmt: off
@pytest.mark.parametrize("map_dims", [
    {"nx": 10, "ny": 5},
    {"nx": 1, "ny": 5},
    {"nx": 2, "ny": 5},
    {"nx": 3, "ny": 5},
    {"nx": 10, "ny": 1},
    {"nx": 10, "ny": 2},
    {"nx": 10, "ny": 3},
])
# fmt: on
def test_fill_quant_fluor_data_dict(map_dims):
    r"""Test for 'fill_quant_fluor_data_dict': testing basic functionality"""
    # Create copy, because the dictionary will be modified
    fluor_standard = copy.deepcopy(_xrf_standard_fluor_sample)

    nx = map_dims["nx"]
    ny = map_dims["ny"]

    img = gen_xrf_map_dict(nx=nx, ny=ny)

    if nx < 3:
        nx_min, nx_max = 0, nx
    else:
        nx_min, nx_max = 1, -1

    if ny < 3:
        ny_min, ny_max = 0, nx
    else:
        ny_min, ny_max = 1, -1

    map_S_K_fluor = np.mean(img["S_K"][ny_min:ny_max, nx_min:nx_max])
    map_Au_M_fluor = np.mean(img["Au_M"][ny_min:ny_max, nx_min:nx_max])
    v_sclr = np.mean(img["sclr"][ny_min:ny_max, nx_min:nx_max])

    # Fill the dictionary, use existing scaler 'sclr'
    fill_quant_fluor_data_dict(fluor_standard, xrf_map_dict=img, scaler_name="sclr")

    npt.assert_almost_equal(
        fluor_standard["element_lines"]["S_K"]["fluorescence"],
        map_S_K_fluor / v_sclr,
        err_msg="Fluorescence of 'S_K' is estimated incorrectly",
    )
    npt.assert_almost_equal(
        fluor_standard["element_lines"]["Au_M"]["fluorescence"],
        map_Au_M_fluor / v_sclr,
        err_msg="Fluorescence of 'Au_M' is estimated incorrectly",
    )
    for eline, param in fluor_standard["element_lines"].items():
        assert (eline in img) or (
            param["fluorescence"] is None
        ), f"Fluorescence line {eline} is not present in the dataset and it was not reset to None"

    # Fill the dictionary, use non-existing scaler 'abc'
    fill_quant_fluor_data_dict(fluor_standard, xrf_map_dict=img, scaler_name="abc")
    npt.assert_almost_equal(
        fluor_standard["element_lines"]["S_K"]["fluorescence"],
        map_S_K_fluor,
        err_msg="Fluorescence of 'S_K' is estimated incorrectly",
    )

    # Fill the dictionary, don't use a scaler (set to None)
    fill_quant_fluor_data_dict(fluor_standard, xrf_map_dict=img, scaler_name=None)
    npt.assert_almost_equal(
        fluor_standard["element_lines"]["S_K"]["fluorescence"],
        map_S_K_fluor,
        err_msg="Fluorescence of 'S_K' is estimated incorrectly",
    )


def test_prune_quant_fluor_data_dict():
    r"""Test for ``prune_quant_fluor_data_dict``: basic test of functionality"""

    # Create copy, because the dictionary will be modified
    fluor_standard = copy.deepcopy(_xrf_standard_fluor_sample)

    elines_not_none = []
    for eline, info in fluor_standard["element_lines"].items():
        set_to_none = np.random.rand() < 0.5  # 50% chance
        if set_to_none:
            info["fluorescence"] = None
        # For some emission lines, fluorescence could already be set to None
        if info["fluorescence"] is not None:
            elines_not_none.append(eline)

    fluor_standard_pruned = prune_quant_fluor_data_dict(fluor_standard)
    for eline, info in fluor_standard_pruned["element_lines"].items():
        assert eline in elines_not_none, f"Emission line {eline} should have been removed from the dictionary"
        assert info["fluorescence"] is not None, f"Emission line {eline} has 'fluorescence' set to None"


def test_set_quant_fluor_data_dict_optional_1():
    r"""
    Tests for 'set_quant_fluor_data_dict_optional': basic test of functionality.
    Successful tests.
    """

    # Create copy, because the dictionary will be modified
    fluor_standard = copy.deepcopy(_xrf_standard_fluor_sample)

    scan_id = 45378  # scan_id has to be int or a string representing int
    scan_uid = "abc-12345"  # Just some string, format is not checked

    set_quant_fluor_data_dict_optional(fluor_standard, scan_id=scan_id, scan_uid=scan_uid)

    # Check if scan_id and scan_uid are set
    assert fluor_standard["source_scan_id"] == scan_id, "Scan ID is set incorrectly"
    assert fluor_standard["source_scan_uid"] == scan_uid, "Scan UID is set incorrectly"

    # Check if time is set correctly
    t = fluor_standard["creation_time_local"]
    assert t is not None, "Time is not set"

    t = convert_time_from_nexus_string(t)
    t = ttime.mktime(t)
    t_current = ttime.mktime(ttime.localtime())

    # 5 seconds is more than sufficient to complete this test
    assert abs(t_current - t) < 5.0, "Time is set incorrectly"

    # Test if sending Scan ID as a string works
    scan_id2 = 45346
    scan_id2_str = f"{scan_id2}"
    set_quant_fluor_data_dict_optional(fluor_standard, scan_id=scan_id2_str)
    assert fluor_standard["source_scan_id"] == scan_id2, "Scan ID is set incorrectly"


def test_set_quant_fluor_data_dict_optional_2():
    r"""
    Tests for 'set_quant_fluor_data_dict_optional'.
    Failing tests.
    """

    # Create copy, because the dictionary will be modified
    fluor_standard = copy.deepcopy(_xrf_standard_fluor_sample)

    # Scan ID is a string, which can not be interpreted as int
    with pytest.raises(RuntimeError, match="Parameter 'scan_id' must be integer or a string representing integer"):
        set_quant_fluor_data_dict_optional(fluor_standard, scan_id="abc_34g")

    # Scan ID is of wrong type
    with pytest.raises(RuntimeError, match="Parameter 'scan_id' must be integer or a string representing integer"):
        set_quant_fluor_data_dict_optional(fluor_standard, scan_id=[1, 5, 14])

    # Scan UID is of wrong type
    with pytest.raises(RuntimeError, match="Parameter 'scan_uid' must be a string representing scan UID"):
        set_quant_fluor_data_dict_optional(fluor_standard, scan_uid=[1, 5, 14])


def test_set_quant_fluor_data_dict_time():

    # Create copy, because the dictionary will be modified
    fluor_standard = copy.deepcopy(_xrf_standard_fluor_sample)

    set_quant_fluor_data_dict_time(fluor_standard)

    # Check if time is set correctly
    t = fluor_standard["creation_time_local"]
    assert t is not None, "Time is not set"

    t = convert_time_from_nexus_string(t)
    t = ttime.mktime(t)
    t_current = ttime.mktime(ttime.localtime())

    # 5 seconds is more than sufficient to complete this test
    assert abs(t_current - t) < 5.0, "Time is set incorrectly"


# --------------------------------------------------------------------------------------------------------


def test_ParamQuantEstimation_1(tmp_path):
    r"""
    Create the object of the 'ParamQuantEstimation' class. Load and then clear reference data.
    """

    # 'home_dir' is typically '~', but for testing it is set to 'tmp_dir'
    home_dir = tmp_path
    config_dir = ".pyxrf"
    standards_fln = "quantitative_standards.yaml"

    pqe = ParamQuantEstimation(home_dir=home_dir)

    # Verify that the cofig file was created
    file_path = os.path.join(home_dir, config_dir, standards_fln)
    assert os.path.isfile(
        file_path
    ), f"Empty file for user-defined reference standards '{file_path}' was not created"

    # Load reference standards
    pqe.load_standards()
    assert pqe.standards_built_in is not None, "Failed to load built-in standards"
    assert pqe.standards_custom is not None, "Failed to load user-defined standards"

    # Clear loaded standards
    pqe.clear_standards()
    assert pqe.standards_built_in is None, "Failed to clear loaded built-in standards"
    assert pqe.standards_custom is None, "Failed to clear loaded user-defined standards"


def test_ParamQuantEstimation_2(tmp_path):

    standard_data = _standard_data_sample

    # 'home_dir' is typically '~', but for testing it is set to 'tmp_dir'
    home_dir = tmp_path
    config_dir = ".pyxrf"
    standards_fln = "quantitative_standards.yaml"

    # Create the file with user-defined reference definitions
    file_path = os.path.join(home_dir, config_dir, standards_fln)
    save_xrf_standard_yaml_file(file_path, standard_data)

    # Create the object and load references
    pqe = ParamQuantEstimation(home_dir=home_dir)
    pqe.load_standards()
    assert pqe.standards_built_in is not None, "Failed to load built-in standards"
    assert len(pqe.standards_built_in) > 0, "The number of loaded built-in standards is ZERO"
    assert pqe.standards_custom is not None, "Failed to load user-defined standards"
    assert len(pqe.standards_custom) > 0, "The number of loaded user-defined standards is ZERO"

    # Verify if the functions for searching the user-defined standards
    for st in pqe.standards_custom:
        serial = st["serial"]
        assert pqe._find_standard_custom(st), f"Standard {serial} was not found in user-defined list"
        assert not pqe._find_standard_built_in(st), f"Standard {serial} was found in built-in list"
        assert pqe.find_standard(st), f"Standard {serial} was not found"
        assert pqe.find_standard(st["name"], key="name"), f"Failed to find standard {serial} by name"
        assert pqe.find_standard(st["serial"], key="serial"), f"Failed to find standard {serial} by serial number"
        assert pqe.is_standard_custom(st), f"Standard {serial} was not identified as user-defined"
        pqe.set_selected_standard(st)
        assert pqe.standard_selected == st, f"Can't select standard {serial}"

    # Verify if the functions for searching the user-defined standards
    for st in pqe.standards_built_in:
        serial = st["serial"]
        assert not pqe._find_standard_custom(st), f"Standard {serial} was found in user-defined list"
        assert pqe._find_standard_built_in(st), f"Standard {serial} was not found in built-in list"
        assert pqe.find_standard(st), f"Standard {serial} was not found"
        assert pqe.find_standard(st["name"], key="name"), f"Failed to find standard {serial} by name"
        assert pqe.find_standard(st["serial"], key="serial"), f"Failed to find standard {serial} by serial number"
        assert not pqe.is_standard_custom(st), f"Standard {serial} was identified as user-defined"
        pqe.set_selected_standard(st)
        assert pqe.standard_selected == st, f"Can't select standard {serial}"

    # Selecting non-existing standard
    st = {"serial": "09876", "name": "Some name"}  # Some arbitatry dictionary
    st_selected = pqe.set_selected_standard(st)
    assert st_selected == pqe.standard_selected, "Return value of 'set_selected_standard' is incorrect"
    assert st_selected == pqe.standards_custom[0], "Incorrect standard is selected"
    # No argument (standard is None)
    pqe.set_selected_standard()
    assert st_selected == pqe.standards_custom[0], "Incorrect standard is selected"
    # Delete the user-defined standards (this is not normal operation, but it's OK for testing)
    pqe.standards_custom = None
    pqe.set_selected_standard(st)
    assert pqe.standard_selected == pqe.standards_built_in[0], "Incorrect standard is selected"
    pqe.standards_built_in = None
    pqe.set_selected_standard(st)
    assert pqe.standard_selected is None, "Incorrect standard is selected"


def test_ParamQuantEstimation_3(tmp_path):

    standard_data = _standard_data_sample

    # 'home_dir' is typically '~', but for testing it is set to 'tmp_dir'
    home_dir = tmp_path
    config_dir = ".pyxrf"
    standards_fln = "quantitative_standards.yaml"

    # Create the file with user-defined reference definitions
    file_path = os.path.join(home_dir, config_dir, standards_fln)
    save_xrf_standard_yaml_file(file_path, standard_data)

    # Create the object and load references
    pqe = ParamQuantEstimation(home_dir=home_dir)
    pqe.load_standards()

    # Select first 'user-defined' sample (from '_standard_data_sample' list)
    incident_energy = 12.0
    img = gen_xrf_map_dict()

    # Generate and fill fluorescence data dictionary
    pqe.set_selected_standard()
    pqe.gen_fluorescence_data_dict(incident_energy=incident_energy)
    scaler_name = "sclr"
    pqe.fill_fluorescence_data_dict(xrf_map_dict=img, scaler_name=scaler_name)

    # Equivalent transformations using functions that are already tested. The sequence
    #   must give the same result (same functions are called).
    qfdd = get_quant_fluor_data_dict(standard_data[0], incident_energy)
    fill_quant_fluor_data_dict(qfdd, xrf_map_dict=img, scaler_name=scaler_name)

    assert (
        pqe.fluorescence_data_dict == qfdd
    ), "The filled fluorescence data dictionary does not match the expected"

    # Test generation of preview (superficial)
    pview, msg_warnings = pqe.get_fluorescence_data_dict_text_preview()
    # It is expected that the preview will contain 'WARNING:'
    assert len(msg_warnings), "Warning is not found in preview"
    # Disable warnings
    pview, msg_warnings = pqe.get_fluorescence_data_dict_text_preview(enable_warnings=False)
    assert not len(msg_warnings), "Warnings are disabled, but still generated"

    # Test function for setting detector channel name
    pqe.set_detector_channel_in_data_dict(detector_channel="sum")
    assert pqe.fluorescence_data_dict["detector_channel"] == "sum", "Detector channel was not set correctly"

    # Test function for setting distance to sample
    distance_to_sample = 2.5
    pqe.set_distance_to_sample_in_data_dict(distance_to_sample=distance_to_sample)
    assert (
        pqe.fluorescence_data_dict["distance_to_sample"] == distance_to_sample
    ), "Distance-to-sample was not set correctly"

    # Function for setting Scan ID and Scan UID
    scan_id = 65476
    scan_uid = "abcdef-12345678"  # Some string

    qfdd = copy.deepcopy(pqe.fluorescence_data_dict)
    pqe.set_optional_parameters(scan_id=scan_id, scan_uid=scan_uid)
    set_quant_fluor_data_dict_optional(qfdd, scan_id=scan_id, scan_uid=scan_uid)
    # We don't check if time is computed correctly (this was checked at different place)
    #   Time computed at different function calls may be different
    qfdd["creation_time_local"] = pqe.fluorescence_data_dict["creation_time_local"]
    assert pqe.fluorescence_data_dict == qfdd, "Optional parameters are not set correctly"

    # Try generating preview again: there should be no warnings
    pview, msg_warnings = pqe.get_fluorescence_data_dict_text_preview()
    # There should be no warnings in preview (all parameters are set)
    assert not len(msg_warnings), "Preview is expected to contain no warnings"

    #  Test the method 'get_suggested_json_fln'
    fln_suggested = pqe.get_suggested_json_fln()
    assert (
        f"_{pqe.fluorescence_data_dict['serial']}." in fln_suggested
    ), f"Serial of the reference is not found in the suggested file name {fln_suggested}"

    # Test saving calibration data
    file_path = os.path.join(tmp_path, fln_suggested)
    pqe.save_fluorescence_data_dict(file_path=file_path)

    qfdd = load_xrf_quant_fluor_json_file(file_path=file_path)
    assert qfdd == prune_quant_fluor_data_dict(
        pqe.fluorescence_data_dict
    ), "Error occurred while saving and loading calibration data dictionary. Dictionaries don't match"


# ---------------------------------------------------------------------------------------------


def _create_file_with_ref_standards(*, wd):
    r"""
    Create a file with user-defined standards based on ``_standard_data_sample[0]``.
    The file contains the descriptions of 2 standards with identical sets of elements/compounds
    with slightly different densities.

    The created file is placed at the standard default location ``<wd>/.pyxrf/quantiative_standards.yaml``.
    Working directory (typically ``~``) is set to temporary directory for using with PyTest.

    It is expected that the dataset will be processed using incident energy value 12.0 keV.
    """
    # Create artificial dataset based on the same standard
    sd = _standard_data_sample[0]
    file_path = os.path.join(wd, ".pyxrf", "quantitative_standards.yaml")

    standard_data = []
    for n in range(2):
        sd_copy = copy.deepcopy(sd)
        # Make sure that the datasets have different serial numbers
        sd_copy["serial"] += f"{n}"
        sd_copy["name"] = f"Test reference standard #{sd_copy['serial']}"
        d = sd_copy["compounds"]
        # Also introduce some small disturbance to the density values
        total_density = 0
        for c in d:
            d[c] += np.random.rand() * 2 - 1
            total_density += d[c]
        sd_copy["density"] = total_density  # Total density also changes
        standard_data.append(sd_copy)

    save_xrf_standard_yaml_file(file_path=file_path, standard_data=standard_data)

    # Create sets of XRF maps for each simulated 'standard'. The images
    #   will have random fluorescence
    img_list = []
    # Pick two overlapping sets of emission lines
    eline_lists = [("Ni_K", "Ga_K", "S_K"), ("S_K", "Cr_K", "Ni_K")]
    for elist in eline_lists:
        img_list.append(gen_xrf_map_dict(elines=elist))

    return file_path, img_list


def _create_files_with_ref_calib(*, wd, img_list, incident_energy=12.0):
    r"""
    Create reference calibration files based on the reference standards read from
    user-defined configuration file at standard location ``<wd>/.pyxrf/quantiative_standards.yaml``.
    The files are created in the working directory <wd>. The list of file paths is returned
    by the function.

    Parameter ``img_list`` is the list of XRF map dictionaries. The number of elements in the list
    must match the number of reference standards. Incident energy is used to determine active
    emission lines, so it has to be set properly.
    """
    pqe = ParamQuantEstimation(home_dir=wd)
    pqe.load_standards()
    pqe.standards_built_in = None  # Remove all 'built-in' standards
    # Now only the 'user-defined' standards are loaded, which were generated by
    #   'create_file_with_ref_standards' function

    file_paths = []
    for n, st in enumerate(pqe.standards_custom):
        pqe.set_selected_standard(st)
        pqe.gen_fluorescence_data_dict(incident_energy=incident_energy)
        pqe.fill_fluorescence_data_dict(xrf_map_dict=img_list[n], scaler_name="sclr")
        pqe.set_detector_channel_in_data_dict(detector_channel="sum")
        pqe.set_distance_to_sample_in_data_dict(distance_to_sample=2.0)
        pqe.set_optional_parameters(scan_id=f"12345{n}", scan_uid="some-uid-string-{n}")
        fln = pqe.get_suggested_json_fln()
        f_path = os.path.join(wd, fln)
        pqe.save_fluorescence_data_dict(file_path=f_path)
        file_paths.append(f_path)

    return file_paths


def create_ref_calib_data(tmp_path, incident_energy=12.0):
    r"""
    Create the dataset, which contains 2 QA reference calibration files (located at ``tmp_path``)
    and XRF map dictionaries with simulated experimental data for each calibration scan. Those
    dictionaries were used to compute fluorescence values for the emission lines during calibration.

    Returns: ``file_paths`` - list of paths to calibration files,
    ``img_list`` - list of simulated XRF map dictionaries, one dictionary per calibration file.
    """
    # Create file with refernce standards
    _, img_list = _create_file_with_ref_standards(wd=tmp_path)

    # Create files with reference calibration data
    file_paths = _create_files_with_ref_calib(wd=tmp_path, img_list=img_list, incident_energy=incident_energy)

    return file_paths, img_list


def test_ParamQuantitativeAnalysis(tmp_path):

    incident_energy = 12.0
    file_paths, img_list = create_ref_calib_data(tmp_path, incident_energy=incident_energy)

    n_entries = len(file_paths)
    assert n_entries >= 2, "The number of calibration data entries must be >= 2. Test can not be performed."
    fln0 = file_paths[0]

    # Load all QA calibration files
    pqa = ParamQuantitativeAnalysis()
    for fpath in file_paths:
        pqa.load_entry(file_path=fpath)  # Use kwarg

    # Check if the number of calibration entries is equal to the number of loaded files
    #   Also check if the function 'get_n_entries' works
    assert pqa.get_n_entries() == n_entries, "Incorrect number of loaded calibration data entries"

    # Check 'get_file_path_list' function
    assert (
        pqa.get_file_path_list() == file_paths
    ), "The returned list of file paths is not the same as the expected list"

    # Attempt to load calibration file that was already loaded
    #   Should not be loaded, the number of entries should remain unchanged
    pqa.load_entry(fln0)  # Send 'fpath' as arg, not kwarg
    assert len(pqa.calibration_data) == n_entries, "Incorrect number of loaded calibration data entries"
    assert len(pqa.calibration_settings) == n_entries, "Unexpected number of 'calibration_settings' elements"

    # Try removing the calibration data entry (the first)
    pqa.remove_entry(file_path=fln0)
    assert len(pqa.calibration_data) == n_entries - 1, "The result of calibration entry removal is not as expected"
    assert len(pqa.calibration_settings) == n_entries - 1, "Unexpected number of 'calibration_settings' elements"

    # Reload the deleted calibration data entry
    pqa.load_entry(fln0)  # Send 'fpath' as arg, not kwarg
    assert len(pqa.calibration_data) == n_entries, "Incorrect number of loaded calibration data entries"
    assert len(pqa.calibration_settings) == n_entries, "Unexpected number of 'calibration_settings' elements"
    # Make sure that the last element contains the last loaded calibration data entry
    assert (
        pqa.calibration_settings[-1]["file_path"] == fln0
    ), "The last element of 'calibration_settings' list contains wrong calibration data entry"

    # Find the index of calibration data entry (by file name)
    for n, fp in enumerate(file_paths):
        n_expected = n - 1
        if n_expected < 0:
            n_expected += n_entries  # n_entries - the number of configuration entries (files)
        assert (
            pqa.find_entry_index(fp) == n_expected
        ), f"Index of the calibration entry '{fp}' is determined incorrectly"

    # Test preview function
    for fp in file_paths:
        assert pqa.get_entry_text_preview(fp), f"Failed to generate text preview for the calibration entry '{fp}'"

    # Get the emission line list and verify if it matches the expected list (including the order of the lines)
    eline_list_expected = []
    for settings in pqa.calibration_settings:
        for ln in settings["element_lines"].keys():
            if ln not in eline_list_expected:
                eline_list_expected.append(ln)
    pqa.update_emission_line_list()
    assert (
        pqa.active_emission_lines == eline_list_expected
    ), "Generated emission line list does not match the expected list"

    # -----------------------------------------------
    # Check how management of selections of calibration entries for individual emission lines
    #   Note, that the calibration entries have the common set of emission lines

    # We will use only the first 2 entries. Find the common set of lines
    elist0 = list(pqa.calibration_settings[0]["element_lines"].keys())
    elist1 = list(pqa.calibration_settings[1]["element_lines"].keys())
    elist = list(set(elist0).intersection(set(elist1)))

    def set_selection(pqa, elist, selection):
        for ln, sel in zip(elist, selection):
            el_info = pqa.get_eline_info_complete(ln)
            assert len(el_info) >= 2, f"The emission line {ln} must exist in at least 2 entries"
            el_info[0]["eline_settings"]["selected"] = sel
            el_info[1]["eline_settings"]["selected"] = not sel

    def check_selection(pqa, elist, selection):
        # We check the selection status directly
        #   (in part to verify if 'get_eline_info_complete' works correctly)
        for ln, sel in zip(elist, selection):
            selected = pqa.calibration_settings[0]["element_lines"][ln]["selected"]
            selected1 = pqa.calibration_settings[1]["element_lines"][ln]["selected"]
            assert selected == sel, f"Entry 0: selection state of element line '{ln}' incorrect"
            assert selected1 != sel, f"Entry 1: selection state of element line '{ln}' incorrect"

    # This should be run at least once to select all entries of the 0th set
    pqa.update_emission_line_list()

    # All elines of group #0 must be selected by default (we didn't change selections)
    sel_list = [True] * len(elist)
    check_selection(pqa, elist, sel_list)

    # Select all elements of group #1
    sel_list = [False] * len(elist)
    set_selection(pqa, elist, sel_list)
    pqa.update_emission_line_list()
    check_selection(pqa, elist, sel_list)

    # Alternatively check selections using 'is_eline_selected'
    for eline in elist1:
        assert pqa.is_eline_selected(
            eline, file_paths[0]
        ), f"Emission line '{eline}' in the set '{file_paths[0]}' was not selected"
        if eline in elist:
            assert not pqa.is_eline_selected(
                eline, file_paths[1]
            ), f"Emission line '{eline}' in the set '{file_paths[1]}' was selected"
    for eline in elist0:
        if eline not in elist:
            assert pqa.is_eline_selected(
                eline, file_paths[1]
            ), f"Emission line '{eline}' in the set '{file_paths[1]}' was not selected"

    # Change selection of one emission line (test for 'select_eline' function)
    eline = elist[0]  # 'eline' is present in both sets
    pqa.select_eline(eline, file_paths[1])
    assert pqa.is_eline_selected(eline, file_paths[1]), "Emission line is not selected"
    assert not pqa.is_eline_selected(eline, file_paths[0]), "Emission line is selected"
    pqa.select_eline(eline, file_paths[0])
    assert pqa.is_eline_selected(eline, file_paths[0]), "Emission line is not selected"
    assert not pqa.is_eline_selected(eline, file_paths[1]), "Emission line is selected"

    # Select one element in both entries (it is already selected in entry #1
    pqa.calibration_settings[0]["element_lines"][elist[0]]["selected"] = True
    pqa.update_emission_line_list()
    # Now the element line with index [0] is selected in entry #0, but the rest are selected in entry #1
    sel_list[0] = True
    pqa.update_emission_line_list()
    check_selection(pqa, elist, sel_list)

    # Delete current entry #0 (entries are arranged in the order 'fln1', 'fln0')
    fln1 = pqa.calibration_settings[0]["file_path"]
    pqa.remove_entry(fln1)
    # Check only the entry #0 (all lines must be selected
    for ln in elist:
        assert pqa.calibration_settings[0]["element_lines"][ln][
            "selected"
        ], f"Element line {ln} is not selected in calibration entry #0"

    # Reload the entry (now the entries are arranged in the order 'fln0', 'fln1')
    #   Entry #0 should still remain selected for all emission lines
    pqa.load_entry(fln1)  # Send 'fpath' as arg, not kwarg
    sel_list = [True] * len(elist)
    check_selection(pqa, elist, sel_list)

    # -----------------------------------------------------

    # Check function 'get_eline_info_complete'
    el_info = pqa.get_eline_info_complete(elist[0])
    assert (
        el_info[0]["eline_data"] == pqa.calibration_data[0]["element_lines"][elist[0]]
    ), "Returned calibration data does not match the expected data"
    assert (
        el_info[0]["eline_settings"] == pqa.calibration_settings[0]["element_lines"][elist[0]]
    ), "Returned calibration settings do not match the settings dictionary"
    assert (
        el_info[0]["standard_data"] == pqa.calibration_data[0]
    ), "Returned standard data does not match the expected data"
    assert (
        el_info[0]["standard_settings"] == pqa.calibration_settings[0]
    ), "Returned standard settings does not match the settings dictionary"

    # Check function 'get_eline_calibration'
    el_calib = pqa.get_eline_calibration(elist[0])
    assert (
        el_calib["fluorescence"] == pqa.calibration_data[0]["element_lines"][elist[0]]["fluorescence"]
    ), "The returned selected emission line calibration is incorrect"

    pqa.set_experiment_detector_channel("sum")
    assert pqa.experiment_detector_channel == "sum", "Detector channel is set incorrectly"

    pqa.set_experiment_distance_to_sample(2.0)
    assert pqa.experiment_distance_to_sample == 2.0, "Distance-to-sample is set incorrectly"

    pqa.set_experiment_incident_energy(12.0)
    assert pqa.experiment_incident_energy == 12.0, "Incident energy is set incorrectly"

    # -------------------------------------------------------------------------------
    # Check processing function 'apply_quantitative_normalization'. Most of the possible
    #   combinations of parameters are checked for cases when quantitative normalization
    #   is performed, q.n. is skipped and regular normalization is performed and no
    #   normalization is performed.

    # Select the emission line, which is definitely exists in the dataset
    img_dict = img_list[0]
    eline = elist[0]  # The dataset is 'img_dict[eline]'

    # Create a separate dictionary of scalers
    scaler, scaler2 = "sclr", "sclr2"
    scaler_dict = {scaler: img_dict[scaler], scaler2: img_dict[scaler] * 2}

    data_mean, scaler_mean = np.mean(img_dict[eline]), np.mean(scaler_dict[scaler])

    # Calibration parameters (we use it to predict the output
    el_calib = pqa.get_eline_calibration(elist[0])
    calib_fluor, calib_density = el_calib["fluorescence"], el_calib["density"]

    # Successfully perform quantitative normalization
    data_out, is_applied = pqa.apply_quantitative_normalization(
        img_dict[eline],
        scaler_dict=scaler_dict,
        scaler_name_default=scaler2,
        data_name=eline,
        name_not_scalable=None,
    )
    npt.assert_almost_equal(
        np.mean(data_out),
        data_mean / scaler_mean * calib_density / calib_fluor,
        err_msg="The value of normalized map is different from expected",
    )
    assert is_applied, "Quantitative normalization was not applied"

    # Ignore correction for the distance to sample
    pqa.set_experiment_distance_to_sample(0.0)
    # Successfully perform quantitative normalization
    data_out, is_applied = pqa.apply_quantitative_normalization(
        img_dict[eline],
        scaler_dict=scaler_dict,
        scaler_name_default=scaler2,
        data_name=eline,
        name_not_scalable=None,
    )
    npt.assert_almost_equal(
        np.mean(data_out),
        data_mean / scaler_mean * calib_density / calib_fluor,
        err_msg="The value of normalized map is different from expected",
    )
    assert is_applied, "Quantitative normalization was not applied"

    # Increase distance-to-sample by the factor of 2
    pqa.set_experiment_distance_to_sample(4.0)
    # Successfully perform quantitative normalization
    data_out, is_applied = pqa.apply_quantitative_normalization(
        img_dict[eline],
        scaler_dict=scaler_dict,
        scaler_name_default=scaler2,
        data_name=eline,
        name_not_scalable=None,
    )
    npt.assert_almost_equal(
        np.mean(data_out),
        data_mean / scaler_mean * calib_density / calib_fluor * 4.0,
        err_msg="The value of normalized map is different from expected",
    )
    assert is_applied, "Quantitative normalization was not applied"

    # Skip quantitative normalization (normalize by 'scaler2') because eline 'data_name'
    #   does not have loaded calibration data associated with it
    data_out, is_applied = pqa.apply_quantitative_normalization(
        img_dict[eline],
        scaler_dict=scaler_dict,
        scaler_name_default=scaler2,
        data_name="Cu_K",
        name_not_scalable=None,
    )
    npt.assert_almost_equal(
        np.mean(data_out),
        data_mean / (scaler_mean * 2),
        err_msg="The value of normalized map is different from expected",
    )
    assert not is_applied, "Quantitative normalization was applied"

    # Skip quantitative normalization (normalize by 'scaler2') because scaler is not
    #   in the scaler dictionary
    scaler_dict2 = scaler_dict.copy()  # The new scaler dict contains only scaler 'sclr2'
    del scaler_dict2["sclr"]
    data_out, is_applied = pqa.apply_quantitative_normalization(
        img_dict[eline],
        scaler_dict=scaler_dict2,
        scaler_name_default=scaler2,
        data_name=eline,
        name_not_scalable=None,
    )
    npt.assert_almost_equal(
        np.mean(data_out),
        data_mean / (scaler_mean * 2),
        err_msg="The value of normalized map is different from expected",
    )
    assert not is_applied, "Quantitative normalization was applied"

    # Skip normalization completely because the scaler name is invalid (non-existing scaler)
    data_out, is_applied = pqa.apply_quantitative_normalization(
        img_dict[eline],
        scaler_dict=scaler_dict,
        scaler_name_default="sclr3",
        data_name="Cu_K",  # Non-existing scaler
        name_not_scalable=["sclr2", "sclr"],
    )
    assert data_out is img_dict[eline], "Function output is not a reference to the input data"
    assert not is_applied, "Quantitative normalization was applied"

    # Skip normalization completely because eline 'data_name'
    #   is in the list of non-scalable names
    data_out, is_applied = pqa.apply_quantitative_normalization(
        img_dict[eline],
        scaler_dict=scaler_dict,
        scaler_name_default=scaler2,
        data_name=eline,
        name_not_scalable=["sclr2", eline, "sclr"],
    )
    assert data_out is img_dict[eline], "Function output is not a reference to the input data"
    assert not is_applied, "Quantitative normalization was applied"

    # ----------------------------------------------------------------------



from qtpy.QtWidgets import QWidget, QLabel, QVBoxLayout, QHBoxLayout, QComboBox
from qtpy.QtCore import Slot

from .useful_widgets import RangeManager, set_tooltip, global_gui_variables
from ..model.lineplot import MapTypes, MapAxesUnits

from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar,
)

import logging

logger = logging.getLogger(__name__)


class PreviewPlotCount(QWidget):
    def __init__(self, *, gpc, gui_vars):
        super().__init__()

        # Global processing classes
        self.gpc = gpc
        # Global GUI variables (used for control of GUI state)
        self.gui_vars = gui_vars

        self.cb_color_scheme = QComboBox()
        # TODO: make color schemes global
        self._color_schemes = ("viridis", "jet", "bone", "gray", "Oranges", "hot")
        self.cb_color_scheme.addItems(self._color_schemes)
        self.cb_color_scheme.currentIndexChanged.connect(self.cb_color_scheme_current_index_changed)

        self.combo_linear_log = QComboBox()
        self.combo_linear_log.addItems(["Linear", "Log"])
        self.combo_linear_log.currentIndexChanged.connect(self.combo_linear_log_current_index_changed)

        self.combo_pixels_positions = QComboBox()
        self.combo_pixels_positions.addItems(["Pixels", "Positions"])
        self.combo_pixels_positions.currentIndexChanged.connect(self.combo_pixels_positions_current_index_changed)

        self.range = RangeManager(add_sliders=True)
        self.range.setMaximumWidth(200)
        self.range.selection_changed.connect(self.range_selection_changed)

        self.mpl_canvas = FigureCanvas(self.gpc.plot_model._fig_maps)
        self.mpl_toolbar = NavigationToolbar(self.mpl_canvas, self)

        # Keep layout without change when canvas is hidden (invisible)
        sp_retain = self.mpl_canvas.sizePolicy()
        sp_retain.setRetainSizeWhenHidden(True)
        self.mpl_canvas.setSizePolicy(sp_retain)

        vbox = QVBoxLayout()

        hbox = QHBoxLayout()
        hbox.addWidget(self.cb_color_scheme)
        hbox.addWidget(self.combo_linear_log)
        hbox.addWidget(self.combo_pixels_positions)
        hbox.addStretch(1)
        hbox.addWidget(QLabel("Range (counts):"))
        hbox.addWidget(self.range)
        vbox.addLayout(hbox)

        vbox.addWidget(self.mpl_toolbar)
        vbox.addWidget(self.mpl_canvas)

        self.setLayout(vbox)

        self._set_tooltips()

    def cb_color_scheme_current_index_changed(self, index):
        logger.debug(f"Color scheme is changed to {self._color_schemes[index]}")
        self.gpc.set_preview_map_color_scheme(self._color_schemes[index])
        self.gpc.update_preview_total_count_map()

    def combo_linear_log_current_index_changed(self, index):
        logger.debug(f"Map type is changed to {MapTypes(index)}")
        self.gpc.set_preview_map_type(MapTypes(index))
        self.gpc.update_preview_total_count_map()

    def combo_pixels_positions_current_index_changed(self, index):
        logger.debug(f"Map axes are changed to {MapAxesUnits(index)}")
        self.gpc.set_preview_map_axes_units(MapAxesUnits(index))
        self.gpc.update_preview_total_count_map()

    def range_selection_changed(self, sel_low, sel_high):
        logger.debug(f"Range selection is changed to ({sel_low:.10g}, {sel_high:.10g})")
        self.gpc.set_preview_map_range(low=sel_low, high=sel_high)
        self.gpc.update_preview_total_count_map()

    @Slot(str)
    def update_map_range(self, mode):
        if mode not in ("reset", "update", "expand", "clear"):
            logger.error(f"PreviewPlotCount.update_map_range: incorrect mode '{mode}'")
            return
        range_low, range_high = self.gpc.get_dataset_preview_count_map_range()

        if (range_low is None) or (range_high is None) or mode == "clear":
            range_low, range_high = 0.0, 1.0  # No datasets are available
            self.range.set_range(range_low, range_high)
            self.range.reset()
        elif mode == "expand":
            # The range may be only expanded. Keep the selection.
            range_low_old, range_high_old = self.range.get_range()
            # sel_low, sel_high = self.range.get_selection()
            range_low = min(range_low, range_low_old)
            range_high = max(range_high, range_high_old)
            self.range.set_range(range_low, range_high)
            # self.range.set_selection(value_low=sel_low, value_high=sel_high)
        elif mode == "update":
            # The selection will expand proportionally to the change in range.
            self.range.set_range(range_low, range_high)
        elif mode == "reset":
            self.range.set_range(range_low, range_high)
            self.range.reset()

        range_low = range_low if range_low is not None else 0
        range_high = range_high if range_high is not None else 0
        self.gpc.set_preview_map_range(low=range_low, high=range_high)

        logger.debug(f"Total Count Preview range is updated: mode='{mode}' range=({range_low}, {range_high})")

    def _set_tooltips(self):
        set_tooltip(self.cb_color_scheme, "Select <b>color scheme</b> for the plotted maps.")
        set_tooltip(
            self.range,
            "<b>Lower and upper limits</b> for the displayed range of intensities. The pixels with "
            "intensities outside the range are <b>clipped</b>.",
        )

    def update_widget_state(self, condition=None):
        if condition == "tooltips":
            self._set_tooltips()
        self.mpl_toolbar.setVisible(self.gui_vars["show_matplotlib_toolbar"])

        # Hide Matplotlib canvas during computations
        state_compute = global_gui_variables["gui_state"]["running_computations"]
        self.mpl_canvas.setVisible(not state_compute)



from __future__ import absolute_import

import os
import copy
import math
from ..model.fileio import FileIOModel
from ..model.lineplot import LinePlotModel
from ..model.parameters import ParamModel, save_as, fit_strategy_list, bound_options
from ..model.draw_image import DrawImageAdvanced
from ..model.draw_image_rgb import DrawImageRGB
from ..model.fit_spectrum import Fit1D, get_cs
from ..model.roi_model import ROIModel
from ..model.param_data import param_data

from ..core.xrf_utils import get_eline_energy

import logging

logger = logging.getLogger(__name__)


class GlobalProcessingClasses:
    def __init__(self):
        self.defaults = None
        self.io_model = None
        self.param_model = None
        self.plot_model = None
        self.fit_model = None
        self.roi_model = None
        self.img_model_adv = None
        self.img_model_rgb = None

    def _get_defaults(self):
        # Set working directory to current working directory (if PyXRF is started from shell)
        working_directory = os.getcwd()
        logger.info(f"Starting PyXRF in the current working directory '{working_directory}'")
        default_parameters = param_data
        return working_directory, default_parameters

    def initialize(self):
        """
        Run the sequence of actions needed to initialize PyXRF modules.
        """
        working_directory, default_parameters = self._get_defaults()
        self.io_model = FileIOModel(working_directory=working_directory)
        self.param_model = ParamModel(default_parameters=default_parameters, io_model=self.io_model)
        self.plot_model = LinePlotModel(param_model=self.param_model, io_model=self.io_model)
        self.fit_model = Fit1D(
            param_model=self.param_model, io_model=self.io_model, working_directory=working_directory
        )
        self.roi_model = ROIModel(param_model=self.param_model, io_model=self.io_model)
        self.img_model_adv = DrawImageAdvanced(io_model=self.io_model)
        self.img_model_rgb = DrawImageRGB(io_model=self.io_model, img_model_adv=self.img_model_adv)

        # Initialization needed to eliminate program crash
        self.plot_model.roi_dict = self.roi_model.roi_dict

        # send working directory changes to different models
        self.io_model.observe("working_directory", self.fit_model.result_folder_changed)
        self.io_model.observe("working_directory", self.roi_model.result_folder_changed)
        self.io_model.observe("selected_file_name", self.fit_model.data_title_update)
        self.io_model.observe("selected_file_name", self.plot_model.exp_label_update)
        self.io_model.observe("selected_file_name", self.roi_model.data_title_update)

        # send the same file to fit model, as fitting results need to be saved
        self.io_model.observe("file_name", self.fit_model.filename_update)
        self.io_model.observe("file_name", self.plot_model.plot_exp_data_update)
        self.io_model.observe("file_name", self.roi_model.filename_update)
        self.io_model.observe("runid", self.fit_model.runid_update)

        # Perform updates when 'io_model.data' is changed (no data is passed)
        self.io_model.observe("data", self.plot_model.exp_data_update)

        # send img dict to img_model for visualization
        self.io_model.observe("img_dict_is_updated", self.fit_model.img_dict_updated)
        self.io_model.observe("img_dict_is_updated", self.img_model_adv.img_dict_updated)
        self.io_model.observe("img_dict_is_updated", self.img_model_rgb.img_dict_updated)

        self.io_model.observe("incident_energy_set", self.plot_model.set_incident_energy)
        self.io_model.observe("incident_energy_set", self.img_model_adv.set_incident_energy)

        self.img_model_adv.observe("scaler_name_index", self.fit_model.scaler_index_update)

        self.img_model_adv.observe("dict_to_plot", self.fit_model.dict_to_plot_update)
        self.img_model_adv.observe("img_title", self.fit_model.img_title_update)

        self.param_model.observe("energy_bound_high_buf", self.plot_model.energy_bound_high_update)
        self.param_model.observe("energy_bound_low_buf", self.plot_model.energy_bound_low_update)

        logger.info("pyxrf started.")

    def add_parameters_changed_cb(self, cb):
        """
        Add callback to the list of callback function that are called after parameters are updated.
        """
        self.param_model.add_parameters_changed_cb(cb)

    def remove_parameters_changed_cb(self, cb):
        """
        Remove reference from the list of callback functions.
        """
        self.param_model.remove_parameters_changed_cb(cb)

    def fitting_parameters_changed(self):
        self.param_model.parameters_changed()

    def get_window_title(self):
        """
        Returns current title of the Main Window of the application.
        """
        return self.io_model.window_title

    def is_databroker_available(self):
        return self.io_model.is_databroker_available()

    def open_data_file(self, file_path):

        self.io_model.data_ready = False
        # only load one file
        # 'temp' is used to reload the same file, otherwise file_name will not update
        self.io_model.file_name = "temp"
        f_dir, f_name = os.path.split(file_path)
        self.io_model.working_directory = f_dir

        def _update_data():
            self.fit_model.fit_img = {}  # clear dict in fitmodel to rm old results
            # This will draw empty (hidden) preview plot, since no channels are selected.
            self.plot_model.update_preview_spectrum_plot()
            self.plot_model.update_total_count_map_preview(new_plot=True)

        # The following statement initiates file loading. It may raise exceptions
        try:
            self.io_model.file_name = f_name

        except Exception:
            _update_data()
            self.fitting_parameters_changed()

            # For plotting purposes, otherwise plot will not update
            self.plot_model.plot_exp_opt = False

            self.plot_model.show_fit_opt = False

            logger.info(f"Failed to load the file '{f_name}'.")
            # Clear file name or scan id from window title. This does not update
            #   the displayed title.
            self.io_model.window_title_clear()
            raise
        else:
            _update_data()
            self.fitting_parameters_changed()

            # For plotting purposes, otherwise plot will not update
            try:
                # TODO: permanent fix for the following issue is needed
                # Suppress Matplotlib exception.
                #   draw_glyph_to_bitmap: glyph num is out of range
                # This is a temporary fix, proper debugging is needed so that try..except
                #   could be removed
                self.plot_model.plot_exp_opt = False
            except Exception as ex:
                logger.error(f"Exception was raised while removing experimental data plot: {str(ex)}")
            try:
                self.plot_model.plot_exp_opt = True
            except Exception as ex:
                logger.error(f"Exception was raised while plotting experimental data: {str(ex)}")
            try:
                self.plot_model.show_fit_opt = False
            except Exception as ex:
                logger.error(f"Exception was raised while removing fitted data plot: {str(ex)}")
            try:
                self.plot_model.show_fit_opt = True
            except Exception as ex:
                logger.error(f"Exception was raised while plotting fitted data: {str(ex)}")

            # Change window title (include file name). This does not update the visible title,
            #   only the text attribute of 'io_model' class.
            self.io_model.window_title_set_file_name(f_name)
            print("======== Set window title")

            if not self.io_model.incident_energy_available:
                msg = (
                    "Incident energy is not available in scan metadata and must be set manually. "
                    "Incident energy may be set by changing 'Incident energy, keV' parameter "
                    "in the dialog boxes opened using 'Find Automatically...' ('Find Elements "
                    "in sample' or 'General...' ('General Settings for Fitting Alogirthm') "
                    "buttons in 'Model' tab."
                )
            else:
                msg = ""

            logger.info(f"Loading of the file '{file_path}' is completed.")

            return msg

    def load_run_from_db(self, run_id_uid):
        """
        Parameters
        ----------
        run_id_uid: int or str
            int - run ID (e.g. -1, 45326), str - run UID
        """

        # Copy value to 'runid' field ('runid' field should not change
        #   during processing unless new data is loaded)
        self.io_model.data_ready = False

        def _update_data():
            self.fit_model.fit_img = {}  # clear dict in fitmodel to rm old results
            # This will draw empty (hidden) preview plot, since no channels are selected.
            self.plot_model.update_preview_spectrum_plot()
            self.plot_model.update_total_count_map_preview(new_plot=True)

        try:
            self.io_model.load_data_runid(run_id_uid)

        except Exception as ex:
            _update_data()
            self.fitting_parameters_changed()

            # For plotting purposes, otherwise plot will not update
            self.plot_model.plot_exp_opt = False
            self.plot_model.show_fit_opt = False

            logger.info(f"Failed to load the run #{run_id_uid}.")
            logger.exception(ex)

            # Clear file name or scan id from window title. This does not update
            #   the displayed title.
            self.io_model.window_title_clear()
            raise

        else:
            _update_data()
            self.fitting_parameters_changed()

            # For plotting purposes, otherwise plot will not update
            self.plot_model.plot_exp_opt = False
            self.plot_model.plot_exp_opt = True
            self.plot_model.show_fit_opt = False
            self.plot_model.show_fit_opt = True

            # Change window title (include file name). This does not update the visible title,
            #   only the text attribute of 'io_model' class.
            self.io_model.window_title_set_run_id(self.io_model.runid)

            file_name = self.io_model.file_name
            msg = ""

            logger.info("Loading of the run is completed")

            return msg, file_name

    def select_preview_dataset(self, *, dset_name, is_visible):
        """
        Select datasets for preview. A dataset may be selected or deselected by
        providing the name of the dataset (a key in `io_model.datasets` dictionary)
        and `is_visible` flag. Multiple datasets may be visualized by selecting the
        datasets one by one.

        Parameters
        ----------
        dset_name: str
            dataset name (a key in `io_model.datasets` dictionary)
        is_visible: bool
            True - show the dataset, False - hide the dataset
        """
        self.io_model.data_sets[dset_name].selected_for_preview = True if is_visible else False
        self.io_model.update_data_set_buffers()
        self.plot_model.update_preview_spectrum_plot()
        self.plot_model.update_total_count_map_preview()

    def get_current_working_directory(self):
        """
        Return current working directory (defined in 'io_model')
        """
        return self.io_model.working_directory

    def set_current_working_directory(self, working_directory):
        """
        Sets current working directory ('io_model')
        """
        self.io_model.working_directory = working_directory

    def get_load_each_channel(self):
        """
        Get the value of `load_each_channel` flag, which indicates if data from
        all detector channels will be loaded.
        """
        return self.io_model.load_each_channel

    def set_load_each_channel(self, load_each_channel):
        """
        Get the value of `load_each_channel` flag, which indicates if data from
        all detector channels will be loaded.
        """
        self.io_model.load_each_channel = load_each_channel

    def get_file_channel_list(self):
        """
        Get the list of channels in the currently loaded file (or run)
        """
        return self.io_model.file_channel_list

    def is_dset_item_selected_for_preview(self, item):
        """
        Check if dataset item is selected for preview.
        """
        return bool(self.io_model.data_sets[item].selected_for_preview)

    def get_loaded_file_name(self):
        """
        Returns the name of currently loaded file.
        """
        return self.io_model.file_name

    def is_scan_metadata_available(self):
        """
        Returns the flag that indicates if scan metadata is loaded
        """
        return bool(self.io_model.scan_metadata_available)

    def get_formatted_metadata(self):
        """
        Returns formatted metadata (for displaying or printing)
        """
        return self.io_model.scan_metadata.get_formatted_output()

    def get_metadata_scan_id(self):
        """
        Reads Run ID from metadata. Check if metadata exists using `is_scan_metadata_available()`
        before calling this function.
        """
        if self.is_scan_metadata_available():
            scan_id = self.io_model.scan_metadata["scan_id"]
        else:
            scan_id = 0  # This is mostly for preventing crashes.
        return scan_id

    def get_metadata_scan_uid(self):
        """
        Reads Run UID from metadata. Check if metadata exists using `is_scan_metadata_available()`
        before calling this function.
        """
        if self.is_scan_metadata_available():
            scan_uid = self.io_model.scan_metadata["scan_uid"]
        else:
            scan_uid = ""  # This is mostly for preventing crashes.
        return scan_uid

    def get_current_run_id(self):
        """
        Returns value of currently set Run ID (in `io_model` class)
        """
        return self.io_model.runid

    def is_xrf_maps_available(self):
        """
        Returns the flag that indicates if XRF maps are currently available (loaded or computed).
        """
        return bool(self.io_model.is_xrf_maps_available())

    def set_data_channel(self, channel_index):
        self.io_model.file_opt = channel_index

        # For plotting purposes, otherwise plot will not update
        self.plot_model.plot_exp_opt = False
        self.plot_model.plot_exp_opt = True
        self.plot_model.show_fit_opt = False
        self.plot_model.show_fit_opt = True

    def is_roi_selection_active(self):
        """
        Returns the flag that indicates if spatial ROI is activated.
        """
        return bool(self.io_model.roi_selection_active)

    def set_roi_selection_active(self, active):
        """
        Sets the flag that indicates if spatial ROI is activated.
        """
        self.io_model.roi_selection_active = active

    def get_preview_spatial_roi(self):
        """
        Returns currently selected spatial ROI (for preview)
        """
        spatial_roi = {
            "row_start": self.io_model.roi_row_start,
            "col_start": self.io_model.roi_col_start,
            "row_end": self.io_model.roi_row_end,
            "col_end": self.io_model.roi_col_end,
        }
        return spatial_roi

    def set_preview_spatial_roi(self, spatial_roi):
        """
        Set spatial ROI (for preview)
        """
        self.io_model.roi_row_start = spatial_roi["row_start"]
        self.io_model.roi_col_start = spatial_roi["col_start"]
        self.io_model.roi_row_end = spatial_roi["row_end"]
        self.io_model.roi_col_end = spatial_roi["col_end"]

    def is_mask_selection_active(self):
        """
        Returns the flag that indicates if mask for ROI selection is activated.
        """
        return bool(self.io_model.mask_active)

    def set_mask_selection_active(self, active):
        """
        Sets the flag that indicates if mask for ROI selection is activated.
        """
        self.io_model.mask_active = active

    def get_mask_selection_file_path(self):
        """
        Returns currently selected path to mask file (ROI selection)
        """
        return self.io_model.mask_file_path

    def set_mask_selection_file_path(self, file_path):
        """
        Sets path to mask file (ROI selection). It also sets mask name.
        """
        self.io_model.mask_file_path = file_path
        # At this point the mask name is just the file name
        self.io_model.mask_name = os.path.split(file_path)[-1]

    def apply_mask_to_datasets(self):
        self.io_model.apply_mask_to_datasets()
        self.plot_model.update_preview_spectrum_plot()

    # ==========================================================================
    #        The following methods are used by the preview widgets

    def get_preview_plot_type(self):
        return self.plot_model.plot_type_preview.value

    def set_preview_plot_type(self, plot_type):
        self.plot_model.plot_type_preview = plot_type

    def get_preview_energy_range(self):
        return self.plot_model.energy_range_preview

    def set_preview_energy_range(self, energy_range):
        self.plot_model.energy_range_preview = energy_range

    def update_preview_spectrum_plot(self):
        self.plot_model.update_preview_spectrum_plot()

    def get_preview_map_color_scheme(self):
        return self.plot_model.map_preview_color_scheme

    def set_preview_map_color_scheme(self, color_scheme):
        self.plot_model.map_preview_color_scheme = color_scheme

    def get_preview_map_type(self):
        return self.plot_model.map_type_preview

    def set_preview_map_type(self, map_type):
        self.plot_model.map_type_preview = map_type

    def get_preview_map_axes_units(self):
        return self.plot_model.map_axes_units_preview

    def set_preview_map_axes_units(self, map_axes_units):
        self.plot_model.map_axes_units_preview = map_axes_units

    def get_preview_map_range(self):
        return self.plot_model.map_preview_range_low, self.plot_model.map_preview_range_high

    def set_preview_map_range(self, low, high):
        self.plot_model.set_map_preview_range(low=low, high=high)

    def get_dataset_preview_count_map_range(self):
        return self.io_model.get_dataset_preview_count_map_range()

    def update_preview_total_count_map(self):
        self.plot_model.update_total_count_map_preview()

    # ==========================================================================
    #          The following methods are used by widgets XRF Maps tab
    #                  to read/write data from model classes

    def get_maps_info_table(self):
        """
        The function builds and returns two tables: a table of value ranges for each map
        in the dataset; a table of limit values that represent selection for the displayed
        values for each map of the dataset.

        Returns
        -------
        range_table: list(list)
            The table is represented as list of lists. Every element of the outer list is
            a table row. Each row has 3 elements: 0 - key (emission line or map name, str),
            1 - lower boundary (float), 2 - upper boundary (float).
        limit_table: list(list)
            The table is represented as list of lists. Every element of the outer list is
            a table row. Each row has 3 elements: 0 - key (emission line or map name, str),
            1 - lower limit value (float), 2 - upper limit value (float).
        """
        # Check if 'range_dict' and 'limit_dict' have the same set of keys
        ks = self.img_model_adv.map_keys
        ks_range = list(self.img_model_adv.range_dict.keys())
        ks_limit = list(self.img_model_adv.limit_dict.keys())
        ks_show = list(self.img_model_adv.stat_dict.keys())
        if set(ks) != set(ks_limit):
            raise RuntimeError(
                "The list of keys in 'limit_dict' is not as expected: "
                f"limit_dict.keys(): {ks_limit} expected: {ks}"
            )
        if set(ks) != set(ks_range):
            raise RuntimeError(
                "The list of keys in 'range_dict' is not as expected: "
                f"range_dict.keys(): {ks_range} expected: {ks}"
            )
        if set(ks) != set(ks_show):
            raise RuntimeError(
                "The list of keys in 'stat_dict' is not as expected: "
                f"stat_dict.keys(): {ks_show} expected: {ks}"
            )

        range_table = []
        limit_table = []
        show_table = []
        for key in ks:
            rng_low = self.img_model_adv.range_dict[key]["low"]
            rng_high = self.img_model_adv.range_dict[key]["high"]
            limit_low_norm = self.img_model_adv.limit_dict[key]["low"]
            limit_high_norm = self.img_model_adv.limit_dict[key]["high"]
            limit_low = rng_low + (rng_high - rng_low) * limit_low_norm / 100.0
            limit_high = rng_low + (rng_high - rng_low) * limit_high_norm / 100.0
            range_table.append([key, rng_low, rng_high])
            limit_table.append([key, limit_low, limit_high])
            show_table.append([key, bool(self.img_model_adv.stat_dict[key])])

        return range_table, limit_table, show_table

    def set_maps_limit_table(self, limit_table, show_table):
        """
        Write updated range limits to 'img_model_adv.limit_dict'. Used by 'Image Wizard'.

        Parameters
        ----------
        limit_table: list(list)
            The table is represented as list of lists. Every element of the outer list is
            a table row. Each row has 3 elements: 0 - key (eline or map name, str),
            1 - low limit (float), 2 - high limit (float).
        """
        # Verify: the keys in both tables must match 'self.img_model_adv.map_keys'
        limit_table_keys = [_[0] for _ in limit_table]
        if set(limit_table_keys) != set(self.img_model_adv.map_keys):
            raise ValueError(
                "GlobalProcessingClasses:set_maps_info_table: keys don't match:"
                f"limit_table has keys {limit_table_keys}, "
                f"original keys {self.img_model_adv.map_keys}"
            )

        show_table_keys = [_[0] for _ in show_table]
        if set(show_table_keys) != set(self.img_model_adv.map_keys):
            raise ValueError(
                "GlobalProcessingClasses:set_maps_info_table: keys don't match:"
                f"show_table has keys {show_table_keys}, "
                f"original keys {self.img_model_adv.map_keys}"
            )

        # Copy limits
        for row in limit_table:
            key, v_low, v_high = row

            rng_low = self.img_model_adv.range_dict[key]["low"]
            rng_high = self.img_model_adv.range_dict[key]["high"]

            v_low_norm = (v_low - rng_low) / (rng_high - rng_low) * 100.0
            v_high_norm = (v_high - rng_low) / (rng_high - rng_low) * 100.0

            self.img_model_adv.limit_dict[key]["low"] = v_low_norm
            self.img_model_adv.limit_dict[key]["high"] = v_high_norm

        # Copy 'show' status (whether the map should be shown)
        for row in show_table:
            key, show_status = row
            self.img_model_adv.stat_dict[key] = bool(show_status)

    def get_maps_dataset_list(self):
        dsets = list(self.io_model.img_dict_keys)
        dset_sel = self.get_maps_selected_dataset()  # The index in the list + 1 (0 - nothing is selected)
        return dsets, dset_sel

    def get_maps_scaler_list(self):
        "Return the list of available scalers for maps"
        scalers = list(self.img_model_adv.scaler_items)
        scaler_sel = self.get_maps_scaler_index()  # The index in the list + 1 (0 - nothing is selected)
        return scalers, scaler_sel

    def get_maps_selected_dataset(self):
        """Returns selected dataset (XRF Maps tab). Index: 0 - no dataset is selected,
        1, 2, ... datasets with index 0, 1, ... is selected"""
        # return int(self.img_model_adv.data_opt)
        return int(self.io_model.img_dict_default_selected_item)

    def set_maps_selected_dataset(self, dataset_index):
        """Select dataset (XRF Maps tab)"""
        # self.img_model_adv.select_dataset(dataset_index)
        self.io_model.select_img_dict_item(dataset_index)

    def get_maps_scaler_index(self):
        return self.img_model_adv.scaler_name_index

    def set_maps_scaler_index(self, scaler_index):
        self.img_model_adv.set_scaler_index(scaler_index)

    def get_maps_quant_norm_enabled(self):
        return bool(self.img_model_adv.quantitative_normalization)

    def set_maps_quant_norm_enabled(self, enable):
        self.img_model_adv.enable_quantitative_normalization(enable)

    def get_maps_scale_opt(self):
        """Returns selected plot type: `Linear` or `Log`"""
        return self.img_model_adv.scale_opt

    def set_maps_scale_opt(self, scale_opt):
        """Set plot type. Allowed values: `Linear` and `Log`"""
        self.img_model_adv.set_scale_opt(scale_opt)

    def get_maps_color_opt(self):
        """Returns selected plot color option: `Linear` or `Log`"""
        return self.img_model_adv.color_opt

    def set_maps_color_opt(self, color_opt):
        """Set plot type. Allowed values: `Linear` and `Log`"""
        self.img_model_adv.set_color_opt(color_opt)

    def get_maps_pixel_or_pos(self):
        """Get the current axes options for the plotted maps. Returned values:
        `Pixes` or `Positions`."""
        values = ["Pixels", "Positions"]
        return values[self.img_model_adv.pixel_or_pos]

    def set_maps_pixel_or_pos(self, pixel_or_pos):
        values = ["Pixels", "Positions"]
        self.img_model_adv.set_pixel_or_pos(values.index(pixel_or_pos))

    def get_maps_show_scatter_plot(self):
        return self.img_model_adv.scatter_show

    def set_maps_show_scatter_plot(self, is_scatter):
        self.img_model_adv.set_plot_scatter(is_scatter)

    def get_maps_grid_interpolate(self):
        return self.img_model_adv.grid_interpolate

    def set_maps_grid_interpolate(self, grid_interpolate):
        self.img_model_adv.set_grid_interpolate(grid_interpolate)

    def compute_map_ranges(self):
        self.img_model_adv.set_low_high_value()

    def redraw_maps(self):
        """Redraw maps in XRF Maps tab"""
        self.img_model_adv.show_image()

    # ==========================================================================
    #          The following methods are used by widgets RGB tab
    #                  to read/write data from model classes

    def get_rgb_maps_info_table(self):
        """
        The function builds and returns two tables: a table of value ranges for each map
        in the dataset; a table of limit values that represent selection for the displayed
        values for each map of the dataset.

        Returns
        -------
        range_table: list(list)
            The table is represented as list of lists. Every element of the outer list is
            a table row. Each row has 3 elements: 0 - key (emission line or map name, str),
            1 - lower boundary (float), 2 - upper boundary (float).
        limit_table: list(list)
            The table is represented as list of lists. Every element of the outer list is
            a table row. Each row has 3 elements: 0 - key (emission line or map name, str),
            1 - lower limit value (float), 2 - upper limit value (float).
        rgb_dict: dict
            dictionary that hold the displayed items: key - color ("red", "green" or "blue"),
            value - selected map represented by the key of `self.img_model_rgb.dict_to_plot`
            or `self.io_model.img_dict[<dataset>]` dictionaries.
        """
        # Check if 'range_dict' and 'limit_dict' have the same set of keys
        ks = self.img_model_rgb.map_keys
        ks_range = list(self.img_model_rgb.range_dict.keys())
        ks_limit = list(self.img_model_rgb.limit_dict.keys())
        rgb_dict = self.img_model_rgb.rgb_dict
        if set(ks) != set(ks_limit):
            raise RuntimeError(
                "The list of keys in 'limit_dict' is not as expected: "
                f"limit_dict.keys(): {ks_limit} expected: {ks}"
            )
        if set(ks) != set(ks_range):
            raise RuntimeError(
                "The list of keys in 'range_dict' is not as expected: "
                f"range_dict.keys(): {ks_range} expected: {ks}"
            )
        if len(rgb_dict) != len(self.img_model_rgb.rgb_keys) or set(rgb_dict.keys()) != set(
            self.img_model_rgb.rgb_keys
        ):
            raise ValueError(
                "GlobalProcessingClasses.get_rgb_maps_info_table(): "
                f"incorrect set of keys in 'rgb_dict': {list(rgb_dict.keys())}, "
                f"expected: {self.img_model_rgb.rgb_keys}"
            )
        for v in rgb_dict.values():
            if (v is not None) and (v not in ks):
                raise ValueError(
                    f"GlobalProcessingClasses.get_rgb_maps_info_table(): Invalid key {v}. Allowed keys: {ks}"
                )

        range_table = []
        limit_table = []
        for key in ks:
            rng_low = self.img_model_rgb.range_dict[key]["low"]
            rng_high = self.img_model_rgb.range_dict[key]["high"]
            limit_low_norm = self.img_model_rgb.limit_dict[key]["low"]
            limit_high_norm = self.img_model_rgb.limit_dict[key]["high"]
            limit_low = rng_low + (rng_high - rng_low) * limit_low_norm / 100.0
            limit_high = rng_low + (rng_high - rng_low) * limit_high_norm / 100.0
            range_table.append([key, rng_low, rng_high])
            limit_table.append([key, limit_low, limit_high])

        return range_table, limit_table, rgb_dict

    def set_rgb_maps_limit_table(self, limit_table, rgb_dict):
        """
        Write updated range limits to 'img_model_adv.limit_dict'. Used by 'Image Wizard'.

        Parameters
        ----------
        limit_table: list(list)
            The table is represented as list of lists. Every element of the outer list is
            a table row. Each row has 3 elements: 0 - key (eline or map name, str),
            1 - low limit (float), 2 - high limit (float).
        rgb_dict: dict
            dictionary that hold the displayed items: key - color ("red", "green" or "blue"),
            value - selected map represented by the key of `self.img_model_rgb.dict_to_plot`
            or `self.io_model.img_dict[<dataset>]` dictionaries.
        """
        # Verify: the keys in both tables must match 'self.img_model_adv.map_keys'
        limit_table_keys = [_[0] for _ in limit_table]
        if set(limit_table_keys) != set(self.img_model_rgb.map_keys):
            raise ValueError(
                "GlobalProcessingClasses:set_maps_info_table: keys don't match:"
                f"limit_table has keys {limit_table_keys}, "
                f"original keys {self.img_model_rgb.map_keys}"
            )

        # Copy limits
        for row in limit_table:
            key, v_low, v_high = row

            rng_low = self.img_model_rgb.range_dict[key]["low"]
            rng_high = self.img_model_rgb.range_dict[key]["high"]

            v_low_norm = (v_low - rng_low) / (rng_high - rng_low) * 100.0
            v_high_norm = (v_high - rng_low) / (rng_high - rng_low) * 100.0

            self.img_model_rgb.limit_dict[key]["low"] = v_low_norm
            self.img_model_rgb.limit_dict[key]["high"] = v_high_norm

        self.img_model_rgb.rgb_dict = rgb_dict.copy()

    def get_rgb_maps_dataset_list(self):
        dsets = list(self.io_model.img_dict_keys)
        dset_sel = self.get_maps_selected_dataset()  # The index in the list + 1 (0 - nothing is selected)
        return dsets, dset_sel

    def get_rgb_maps_scaler_list(self):
        "Return the list of available scalers for maps"
        scalers = list(self.img_model_rgb.scaler_items)
        scaler_sel = self.get_maps_scaler_index()  # The index in the list + 1 (0 - nothing is selected)
        return scalers, scaler_sel

    def get_rgb_maps_selected_dataset(self):
        """Returns selected dataset (XRF Maps tab). Index: 0 - no dataset is selected,
        1, 2, ... datasets with index 0, 1, ... is selected"""
        # return int(self.img_model_rgb.data_opt)
        return int(self.io_model.img_dict_default_selected_item)

    def set_rgb_maps_selected_dataset(self, dataset_index):
        """Select dataset (XRF Maps tab)"""
        # self.img_model_rgb.select_dataset(dataset_index)
        self.io_model.select_img_dict_item(dataset_index)

    def get_rgb_maps_scaler_index(self):
        return self.img_model_rgb.scaler_name_index

    def set_rgb_maps_scaler_index(self, scaler_index):
        self.img_model_rgb.set_scaler_index(scaler_index)

    def get_rgb_maps_pixel_or_pos(self):
        """Get the current axes options for the plotted maps. Returned values:
        `Pixes` or `Positions`."""
        values = ["Pixels", "Positions"]
        return values[self.img_model_rgb.pixel_or_pos]

    def set_rgb_maps_pixel_or_pos(self, pixel_or_pos):
        values = ["Pixels", "Positions"]
        self.img_model_rgb.set_pixel_or_pos(values.index(pixel_or_pos))

    def get_rgb_maps_grid_interpolate(self):
        return self.img_model_rgb.grid_interpolate

    def set_rgb_maps_grid_interpolate(self, grid_interpolate):
        self.img_model_rgb.set_grid_interpolate(grid_interpolate)

    def get_rgb_maps_quant_norm_enabled(self):
        return bool(self.img_model_rgb.quantitative_normalization)

    def set_rgb_maps_quant_norm_enabled(self, enable):
        self.img_model_rgb.enable_quantitative_normalization(enable)

    def compute_rgb_map_ranges(self):
        self.img_model_rgb.set_low_high_value()

    def redraw_rgb_maps(self):
        """Redraw maps in XRF Maps tab"""
        self.img_model_rgb.show_image()

    # ==========================================================================
    #          The following methods are used by widgets related
    #      to loading and management of quantitative calibration data
    def load_quantitative_calibration_data(self, file_path):
        self.img_model_adv.load_quantitative_calibration_data(file_path)

    def get_quant_calibration_data(self):
        return self.img_model_adv.param_quant_analysis.calibration_data

    def get_quant_calibration_settings(self):
        return self.img_model_adv.param_quant_analysis.calibration_settings

    def get_quant_calibration_text_preview(self, file_path):
        return self.img_model_adv.param_quant_analysis.get_entry_text_preview(file_path)

    def quant_calibration_remove_entry(self, file_path):
        return self.img_model_adv.param_quant_analysis.remove_entry(file_path)

    def get_quant_calibration_file_path_list(self):
        return self.img_model_adv.param_quant_analysis.get_file_path_list()

    def get_quant_calibration_is_eline_selected(self, eline, file_path):
        return self.img_model_adv.param_quant_analysis.is_eline_selected(eline, file_path)

    def set_quant_calibration_select_eline(self, eline, file_path):
        self.img_model_adv.param_quant_analysis.select_eline(eline, file_path)

    def get_quant_calibration_distance_to_sample(self):
        return self.img_model_adv.param_quant_analysis.experiment_distance_to_sample

    def set_quant_calibration_distance_to_sample(self, distance_to_sample):
        self.img_model_adv.param_quant_analysis.set_experiment_distance_to_sample(distance_to_sample)

    # ==========================================================================
    #    The following methods are used by Fitting Model tab (pectrum plots)

    def get_line_plot_state(self):
        plot_spectrum = bool(self.plot_model.plot_exp_opt)
        plot_fit = bool(self.plot_model.show_fit_opt)
        return plot_spectrum, plot_fit

    def show_plot_spectrum(self, state):
        self.plot_model.plot_exp_opt = bool(state)

    def show_plot_fit(self, state):
        self.plot_model.show_fit_opt = bool(state)

    def get_plot_fit_energy_range(self):
        return self.plot_model.energy_range_fitting

    def set_plot_fit_energy_range(self, range_name):
        if range_name not in self.plot_model.energy_range_names:
            raise ValueError(
                f"Range name {range_name} is not in the list of allowed "
                f"names {self.plot_model.energy_range_name}"
            )
        self.plot_model.set_energy_range_fitting(range_name)

    def get_plot_fit_linlog(self):
        if self.plot_model.scale_opt == 0:
            return "linlog"
        else:
            return "linear"

    def set_plot_fit_linlog(self, scale):
        """Scale may have values 'linlog' or 'linear'.
        Raises `ValueError` if other value is passed."""
        self.plot_model.scale_opt = ["linlog", "linear"].index(scale)

    def update_plot_fit(self):
        self.plot_model.plot_experiment()

    def get_incident_energy(self):
        return self.plot_model.incident_energy

    def get_escape_peak_params(self):
        plot_escape_peak = bool(self.plot_model.plot_escape_line)
        materials = ["Si", "Ge"]
        try:
            detector_material = materials[self.plot_model.det_materials]
        except Exception:
            detector_material = ""
        return plot_escape_peak, detector_material

    def set_escape_peak_params(self, plot_escape_peak, detector_material):
        materials = ["Si", "Ge"]
        if detector_material not in materials:
            raise ValueError(f"Detector material '{detector_material}' is unknown: {materials}")
        self.plot_model.change_escape_peak_settings(
            1 if plot_escape_peak else 0, materials.index(detector_material)
        )

    # ==========================================================================
    #          The following methods are used by Model tab
    def get_autofind_elements_params(self):
        """Assemble the parameters managed by 'Find Elements in Sample` dialog box."""
        dialog_params = {}
        keys1 = ["e_offset", "e_linear", "e_quadratic", "fwhm_offset", "fwhm_fanoprime", "coherent_sct_energy"]
        for k in keys1:
            dialog_params[k] = {}
            dialog_params[k]["value"] = self.param_model.param_new[k]["value"]
            dialog_params[k]["default"] = self.param_model.param_new[k]["default"]
        keys2 = ["energy_bound_high", "energy_bound_low"]
        for k in keys2:
            dialog_params[k] = {}
            dialog_params[k]["value"] = self.param_model.param_new["non_fitting_values"][k]["value"]
            dialog_params[k]["default"] = self.param_model.param_new["non_fitting_values"][k]["default_value"]

        # Some values were taken from different places
        dialog_params["coherent_sct_energy"]["value"] = self.plot_model.incident_energy

        return dialog_params

    def set_autofind_elements_params(self, dialog_params, *, update_model=True, update_fitting_params=True):
        """Save the parameters changed in 'Find Elements in Sample` dialog box.

        Parameters
        ----------
        dialog_params : dict
            parameters from dialog box
        update_model : boolean
            True - update the spectra based on new parameters. Updating the spectra
            may take long time if model contains a lot of parameters, therefore it
            should not be used for the default model which contains all supported emission
            lines.
        update_fitting_params : bool
            True - execute callback functions after updating fitting parameters

        Returns
        -------
        boolean
            True - selected range or incident energy were changed, False otherwise
        """
        logger.debug("Saving parameters from 'DialogFindElements'")

        # Check if the critical parameters were changed
        if (
            (self.plot_model.incident_energy != dialog_params["coherent_sct_energy"]["value"])
            or (self.param_model.energy_bound_high_buf != dialog_params["energy_bound_high"]["value"])
            or (self.param_model.energy_bound_low_buf != dialog_params["energy_bound_low"]["value"])
            or (self.param_model.param_new["e_offset"]["value"] != dialog_params["e_offset"]["value"])
            or (self.param_model.param_new["e_linear"]["value"] != dialog_params["e_linear"]["value"])
            or (self.param_model.param_new["e_quadratic"]["value"] != dialog_params["e_quadratic"]["value"])
        ):
            return_value = True
        else:
            return_value = False

        keys = ["e_offset", "e_linear", "e_quadratic", "fwhm_offset", "fwhm_fanoprime"]
        for k in keys:
            self.param_model.param_new[k]["value"] = dialog_params[k]["value"]

        self.io_model.incident_energy_set = dialog_params["coherent_sct_energy"]["value"]
        self.param_model.energy_bound_high_buf = dialog_params["energy_bound_high"]["value"]
        self.param_model.energy_bound_low_buf = dialog_params["energy_bound_low"]["value"]

        # Also change incident energy
        self.param_model.param_new["coherent_sct_energy"]["value"] = dialog_params["coherent_sct_energy"]["value"]
        self.param_model.param_new["non_fitting_values"]["energy_bound_low"]["value"] = dialog_params[
            "energy_bound_low"
        ]["value"]
        self.param_model.param_new["non_fitting_values"]["energy_bound_high"]["value"] = dialog_params[
            "energy_bound_high"
        ]["value"]

        if update_model and return_value:
            self.param_model.create_spectrum_from_param_dict(reset=False)

        if update_fitting_params:
            self.fitting_parameters_changed()

        return return_value

    def get_general_fitting_params(self):
        params = dict()

        # Another option is self.param_model.param_new
        source_dict = self.param_model.param_new

        params["max_iterations"] = self.fit_model.fit_num
        params["tolerance"] = self.fit_model.ftol
        params["escape_peak_ratio"] = source_dict["non_fitting_values"]["escape_ratio"]

        params["incident_energy"] = source_dict["coherent_sct_energy"]["value"]
        params["range_low"] = source_dict["non_fitting_values"]["energy_bound_low"]["value"]
        params["range_high"] = source_dict["non_fitting_values"]["energy_bound_high"]["value"]

        params["subtract_baseline_linear"] = self.fit_model.linear_bg
        params["subtract_baseline_snip"] = self.fit_model.use_snip
        params["add_bias"] = self.fit_model.raise_bg

        params["snip_window_size"] = source_dict["non_fitting_values"]["background_width"]

        return params

    def set_general_fitting_params(self, params):

        # Another option is self.param_model.param_new
        dest_dict = self.param_model.param_new

        self.fit_model.fit_num = params["max_iterations"]
        self.fit_model.ftol = params["tolerance"]
        dest_dict["non_fitting_values"]["escape_ratio"] = params["escape_peak_ratio"]

        dest_dict["coherent_sct_energy"]["value"] = params["incident_energy"]
        dest_dict["non_fitting_values"]["energy_bound_low"]["value"] = params["range_low"]
        dest_dict["non_fitting_values"]["energy_bound_high"]["value"] = params["range_high"]
        self.io_model.incident_energy_set = params["incident_energy"]

        self.fit_model.linear_bg = params["subtract_baseline_linear"]
        self.fit_model.use_snip = params["subtract_baseline_snip"]
        self.fit_model.raise_bg = params["add_bias"]

        dest_dict["non_fitting_values"]["background_width"] = params["snip_window_size"]

        self.param_model.create_spectrum_from_param_dict(reset=False)
        self.apply_to_fit()

        self.fitting_parameters_changed()

    def get_fit_strategy_list(self):
        return fit_strategy_list.copy()

    def get_detailed_fitting_params_lines(self):
        # Dictionary of fitting parameters. Create a copy !!!
        param_dict = copy.deepcopy(self.param_model.param_new)
        # Ordered list of emission lines
        eline_list = self.param_model.get_sorted_element_list()
        # Dictionary: emission line -> [list of keys in 'params' dictionary]
        eline_key_dict = dict()
        eline_energy_dict = dict()
        for eline in eline_list:
            key = None
            eline_category = self.get_eline_name_category(eline)
            if eline_category == "eline":
                key = eline[:-1] + eline[-1].lower()
            elif eline_category == "pileup":
                key = f"pileup_{eline.replace('-', '_')}"
            elif eline_category == "userpeak":
                key = eline

            if key:
                e_list = [_ for _ in param_dict.keys() if _.startswith(key)]
            else:
                e_list = []

            energy_list = []
            if eline_category == "eline":
                for key in e_list:
                    el = "_".join(key.split("_")[:2])
                    energy_list.append(get_eline_energy(el))
            elif eline_category == "pileup":
                energy = self.param_model.get_pileup_peak_energy(eline)
                energy_list = [energy] * len(e_list)
            elif eline_category == "userpeak":
                key = eline + "_delta_center"
                energy = param_dict[key]["value"] + 5.0
                energy_list = [energy] * len(e_list)

            eline_key_dict[eline] = e_list
            eline_energy_dict[eline] = energy_list

        # List of other parameters in 'params' dictionary (except non_fitting_values).
        #   Currently all the parameters not related to emission lines are strictly lower-case.
        other_param_list = [_ for _ in param_dict.keys() if (_ == _.lower()) and (_ != "non_fitting_values")]
        other_param_list.sort()
        # eline_list = ["Shared parameters"] + eline_list

        _fit_strategy_list = fit_strategy_list.copy()
        _bound_options = bound_options.copy()

        return {
            "param_dict": param_dict,
            "eline_list": eline_list,
            "eline_key_dict": eline_key_dict,
            "eline_energy_dict": eline_energy_dict,
            "other_param_list": other_param_list,
            "fit_strategy_list": _fit_strategy_list,
            "bound_options": _bound_options,
        }

    def get_detailed_fitting_params_shared(self):
        # Dictionary of fitting parameters. Create a copy !!!
        param_dict = copy.deepcopy(self.param_model.param_new)
        # Ordered list of emission lines
        eline_list = ["Shared parameters"]
        # Dictionary: emission line -> [list of keys in 'params' dictionary]
        eline_key_dict = dict()
        eline_energy_dict = dict()

        # List of other parameters in 'params' dictionary (except non_fitting_values).
        #   Currently all the parameters not related to emission lines are strictly lower-case.
        other_param_list = [_ for _ in param_dict.keys() if (_ == _.lower()) and (_ != "non_fitting_values")]
        other_param_list.sort()

        _fit_strategy_list = fit_strategy_list.copy()
        _bound_options = bound_options.copy()

        return {
            "param_dict": param_dict,
            "eline_list": eline_list,
            "eline_key_dict": eline_key_dict,
            "eline_energy_dict": eline_energy_dict,
            "other_param_list": other_param_list,
            "fit_strategy_list": _fit_strategy_list,
            "bound_options": _bound_options,
        }

    def set_detailed_fitting_params(self, dialog_data):
        param_dict = dialog_data["param_dict"]
        # 'param_dict' is expected to have identical structure as 'param_model.param_new'.
        # We don't want to change the reference to the parameters, so we copy references to
        #   dictionary elements (which are also dictionaries).

        # Check if the energy axis parameters were changed
        e_axis_changed = False
        e_axis_keys = ("e_offset", "e_linear", "e_quadratic")
        for key in e_axis_keys:
            if self.param_model.param_new[key]["value"] != param_dict[key]["value"]:
                e_axis_changed = True
                break

        for key, val in param_dict.items():
            self.param_model.param_new[key] = val

        self.io_model.incident_energy_set = param_dict["coherent_sct_energy"]["value"]

        self.param_model.create_spectrum_from_param_dict(reset=False)
        self.apply_to_fit()

        # This is needed in case the coefficients that define energy axis have changed
        if e_axis_changed:
            self.update_preview_spectrum_plot()

        self.fitting_parameters_changed()

    def get_quant_standard_list(self):
        self.fit_model.param_quant_estimation.clear_standards()
        self.fit_model.param_quant_estimation.load_standards()

        qe_param_built_in = self.fit_model.param_quant_estimation.standards_built_in
        qe_param_custom = self.fit_model.param_quant_estimation.standards_custom

        # The current selection is set to the currently selected standard
        #   held in 'fit_model.qe_standard_selected_copy'
        qe_standard_selected = self.fit_model.qe_standard_selected_copy
        qe_standard_selected = self.fit_model.param_quant_estimation.set_selected_standard(qe_standard_selected)

        return qe_param_built_in, qe_param_custom, qe_standard_selected

    def set_selected_quant_standard(self, selected_standard):
        if selected_standard is not None:
            self.fit_model.param_quant_estimation.set_selected_standard(selected_standard)
            self.fit_model.qe_standard_selected_copy = copy.deepcopy(selected_standard)
        else:
            self.fit_model.param_quant_estimation.clear_standards()
            self.fit_model.qe_standard_selected_copy = None

        self.fitting_parameters_changed()

    def is_quant_standard_custom(self, standard=None):
        return self.fit_model.param_quant_estimation.is_standard_custom(standard)

    def process_peaks_from_quantitative_sample_data(self):
        incident_energy = self.param_model.param_new["coherent_sct_energy"]["value"]
        # Generate the data structure for the results of processing of standard data
        self.fit_model.param_quant_estimation.gen_fluorescence_data_dict(incident_energy)
        # Obtain the list (dictionary) of elemental lines from the generated structure
        elemental_lines = self.fit_model.param_quant_estimation.fluorescence_data_dict["element_lines"]

        self.param_model.find_peak(elemental_lines=elemental_lines.keys())
        self.param_model.EC.order()
        self.param_model.update_name_list()

        self.param_model.EC.turn_on_all()
        self.param_model.data_for_plot()

        # update parameter for fit
        self.param_model.create_full_param()  # Not sure it is necessary
        self.fit_model.apply_default_param()

        # update experimental plots in case the coefficients change
        self.plot_model.plot_experiment()

        self.plot_model.plot_fit(
            self.param_model.prefit_x, self.param_model.total_y, self.param_model.auto_fit_all
        )

        # Update displayed intensity of the selected peak
        self.plot_model.compute_manual_peak_intensity()

        # Show the summed spectrum used for fitting
        self.plot_model.plot_exp_opt = False
        self.plot_model.plot_exp_opt = True
        # For plotting purposes, otherwise plot will not update
        self.plot_model.show_fit_opt = False
        self.plot_model.show_fit_opt = True

        self.fitting_parameters_changed()

    def load_parameters_from_file(self, parameter_file_path, incident_energy_from_param_file=None):

        try:
            self.param_model.read_param_from_file(parameter_file_path)
        except Exception as ex:
            msg = f"Error occurred while reading parameter file: {ex}"
            logger.error(msg)
            raise IOError(msg)
        else:
            # Make a decision, if the incident energy from metadata should be replaced
            #   with the incident energy from the parameter file
            overwrite_metadata_incident_energy = False

            # Incident energy from the parameter file
            param_incident_energy = self.param_model.param_new["coherent_sct_energy"]["value"]

            if self.io_model.incident_energy_available:

                # Incident energy from datafile metadata
                mdata_incident_energy = self.io_model.scan_metadata.get_mono_incident_energy()

                # If two energies have very close values (say 1 eV), then the difference doesn't matter
                #   Consider two energies equal (they probably ARE equal) and use the value
                #   from datafile metadata.
                if not math.isclose(param_incident_energy, mdata_incident_energy, abs_tol=0.001):
                    # TODO: the following text is not properly formatted for QMessageBox
                    #       It's not very important now, but should be resolved in the future.
                    msg = (
                        f"The values of incident energy from data file metadata "
                        f"and parameter file are different.\n"
                        f"Incident energy from metadata: {mdata_incident_energy} keV.\n"
                        f"Incident energy from the loaded parameter file: {param_incident_energy} keV.\n"
                        f"Would you prefer to use the incident energy from the parameter file for processing?"
                    )

                    if incident_energy_from_param_file is None:
                        return False, msg
                    else:
                        overwrite_metadata_incident_energy = bool(incident_energy_from_param_file)

            else:

                # If incident energy is not present in file metadata, then
                #   just load the incident energy from the parameter file.
                overwrite_metadata_incident_energy = True

            if overwrite_metadata_incident_energy:
                logger.info(f"Using incident energy from the parameter file: {param_incident_energy} keV")
            else:
                # Keep the incident energy from the file
                logger.info(f"Using incident energy from the datafile metadata: {mdata_incident_energy} keV")
                incident_energy = round(mdata_incident_energy, 6)
                self.param_model.param_new["coherent_sct_energy"]["value"] = incident_energy
                self.param_model.param_new["non_fitting_values"]["energy_bound_high"]["value"] = (
                    incident_energy + 0.8
                )

            self.fit_model.apply_default_param()

            # update experimental plots
            self.plot_model.plot_experiment()
            self.plot_model.plot_exp_opt = False
            self.plot_model.plot_exp_opt = True

            # calculate profile and plot
            self.fit_model.get_profile()

            # update experimental plot with new calibration values
            self.plot_model.plot_fit(
                self.fit_model.cal_x, self.fit_model.cal_y, self.fit_model.cal_spectrum, self.fit_model.residual
            )

            self.plot_model.plot_exp_opt = False
            self.plot_model.plot_exp_opt = True
            # For plotting purposes, otherwise plot will not update
            self.plot_model.show_fit_opt = False
            self.plot_model.show_fit_opt = True

            # The following statement is necessary mostly to set the correct value of
            #   the upper boundary of the energy range used for emission line search.
            self.plot_model.change_incident_energy(self.param_model.param_new["coherent_sct_energy"]["value"])

            # update parameter for fit
            # self.param_model.create_full_param()
            self.fit_model.apply_default_param()

            # Update displayed intensity of the selected peak
            self.plot_model.compute_manual_peak_intensity()

            self.fitting_parameters_changed()

            return True, ""

    def find_elements_automatically(self):
        logger.debug("Starting automated element search")
        self.param_model.find_peak()

        self.param_model.EC.order()
        self.param_model.update_name_list()

        self.param_model.EC.turn_on_all()

        self.plot_model.plot_exp_opt = True
        self.apply_to_fit()

        self.fitting_parameters_changed()

    def get_full_eline_list(self):
        """Returns full list of supported emission lines."""
        return self.param_model.get_user_peak_list()

    def set_marker_reporter(self, callback):
        self.plot_model.report_marker_state = callback

    def select_eline(self, eline):
        self.param_model.e_name = eline
        n_id = 0  # No element is selected
        if eline != "":
            eline_list = self.get_full_eline_list()
            try:
                n_id = eline_list.index(eline) + 1  # Elements are numbered starting with 1.
            except ValueError:
                pass
        self.plot_model.element_id = n_id

    def set_selected_eline(self, eline):
        """Sets and display the location of current emission line"""
        self.select_eline(eline)

        eline_category = self.get_eline_name_category(eline)
        if eline_category == "userpeak":
            self.fit_model.select_index_by_eline_name(eline)
            # This will delete the peak lines (no peak lines are plotted for user peak)
            self.plot_model.plot_current_eline(eline)
            # Show the marker at the center of the userpeak
            energy, _ = self.param_model.get_selected_eline_energy_fwhm(eline)
            self.plot_model.set_plot_vertical_marker(energy)
        elif eline_category == "pileup":
            # The lines for pileup peak are not displayed as 'plot_model.element_id' is changed.
            #   So call plotting function explicitly
            self.plot_model.plot_current_eline(eline)
            self.plot_model.hide_plot_vertical_marker()
        elif eline_category == "eline":
            self.plot_model.hide_plot_vertical_marker()
        else:
            # This will delete the peak lines
            self.plot_model.plot_current_eline(eline)
            self.plot_model.hide_plot_vertical_marker()

    def get_eline_intensity(self, eline):
        line_category = self.get_eline_name_category(eline)
        if line_category == "eline":
            return self.param_model.add_element_intensity
        elif line_category in ("userpeak", "pileup"):
            return self.param_model.EC.element_dict[eline].maxv
        else:
            return None

    def get_selected_eline_table(self):
        """Returns full list of supported emission lines."""
        eline_names = self.param_model.get_sorted_result_dict_names()
        eline_table = []
        for eline in eline_names:
            sel_status = self.param_model.EC.element_dict[eline].status
            z = self.param_model.EC.element_dict[eline].z
            energy = self.param_model.EC.element_dict[eline].energy
            peak_int = self.param_model.EC.element_dict[eline].maxv
            rel_int = self.param_model.EC.element_dict[eline].norm
            cs = get_cs(eline, self.param_model.param_new["coherent_sct_energy"]["value"])
            row_data = {
                "eline": eline,
                "sel_status": sel_status,
                "z": z,
                "energy": energy,
                "peak_int": peak_int,
                "rel_int": rel_int,
                "cs": cs,
            }
            eline_table.append(row_data)
        return eline_table

    def get_unused_userpeak_name(self):
        eline_names = self.param_model.get_sorted_result_dict_names()
        for n in range(1, 10000):  # Set some arbitrarily high limit to avoid infinite loops
            userpeak_name = f"Userpeak{n}"
            if userpeak_name not in eline_names:
                return userpeak_name

    def get_eline_name_category(self, eline_name):
        return self.param_model.get_eline_name_category(eline_name)

    def set_checked_emission_lines(self, eline_list, eline_checked_list):
        for eline, checked in zip(eline_list, eline_checked_list):
            self.param_model.EC.element_dict[eline].status = checked
            self.param_model.EC.element_dict[eline].stat_copy = self.param_model.EC.element_dict[eline].status

        self.param_model.data_for_plot()

        self.plot_model.plot_fit(
            self.param_model.prefit_x, self.param_model.total_y, self.param_model.auto_fit_all
        )
        # For plotting purposes, otherwise plot will not update
        self.plot_model.show_fit_opt = False
        self.plot_model.show_fit_opt = True

    def get_current_userpeak_energy_fwhm(self):
        return self.param_model.get_selected_eline_energy_fwhm(self.param_model.e_name)

    def generate_pileup_peak_name(self, eline1, eline2):
        return self.param_model.generate_pileup_peak_name(eline1, eline2)

    def get_pileup_peak_energy(self, eline):
        return self.param_model.get_pileup_peak_energy(eline)

    def is_peak_already_selected(self, eline):
        """Verifies if emission line with the same name is already selected"""
        line_list = list(self.param_model.EC.element_dict.keys())
        if eline in line_list:
            return True
        else:
            return False

    def convert_full_eline_name(self, eline, to_upper=True):
        """Convert emission lines to the form used for displayed emission line names"""
        try:
            index = eline.index("_")
            eline_as_list = list(eline)
            if to_upper:
                eline_as_list[index + 1] = eline_as_list[index + 1].upper()
            else:
                eline_as_list[index + 1] = eline_as_list[index + 1].lower()
            eline = "".join(eline_as_list)
        except Exception:
            pass
        return eline

    def get_guessed_pileup_peak_components(self, energy, tolerance=0.05):
        """
        Returns tuple (line1, line2, pileup_energy), that contains the best guess
        for the pileup peak placed `energy`.

        Parameters
        ----------
        energy: float
            Approximate (selected) energy of pileup peak location
        tolerance: float
            Allowed deviation of the sum of component energies from the selected energy, keV

        Returns
        -------
        tuple(str, str, float)
            Component emission lines (such as Ca_ka1, K_ka1 etc) and the energy of
            the resulting pileup peak.
        """
        eline_info = self.param_model.guess_pileup_peak_components(energy=energy, tolerance=tolerance)
        if eline_info is not None:
            e1, e2, energy = eline_info
            eline_info = (
                self.convert_full_eline_name(e1, to_upper=True),
                self.convert_full_eline_name(e2, to_upper=True),
                energy,
            )
        return eline_info

    def get_selected_energy_range(self):
        """
        Returns the energy range selected for processing. No peaks are supposed to be selected
        outside this range.

        Returns
        -------
        tuple(float, float)
            Selected range (low, high) in keV
        """
        e_low = self.param_model.param_new["non_fitting_values"]["energy_bound_low"]["value"]
        e_high = self.param_model.param_new["non_fitting_values"]["energy_bound_high"]["value"]
        return e_low, e_high

    def add_peak_manual(self, eline):
        """Manually add a peak (emission line) using 'Add' button"""
        self.select_eline(eline)

        # Verify if we are adding a userpeak
        is_userpeak = self.get_eline_name_category(eline) == "userpeak"

        # The following set of conditions is not complete, but sufficient
        if self.param_model.x0 is None or self.param_model.y0 is None:
            err_msg = "Experimental data is not loaded or initial\nspectrum fitting is not performed"
            raise RuntimeError(err_msg)

        elif is_userpeak and (not self.plot_model.vertical_marker_is_visible):
            err_msg = (
                "Select position of userpeak by clicking on the spectrum plot.\n"
                "Note: plot toolbar options, such as Pan and Zoom, \n"
                "must be turned off before selecting the position."
            )
            raise RuntimeError(err_msg)

        else:
            try:
                # Energy is used for user-defined peaks only
                energy, _ = self.get_suggested_manual_peak_energy()

                self.param_model.add_peak_manual(userpeak_center=energy)
                self.apply_to_fit()
                self.fit_model.select_index_by_eline_name(eline)
            except Exception as ex:
                raise RuntimeError(str(ex))

        self.fitting_parameters_changed()

    def remove_peak_manual(self, eline):
        """Manually add a peak (emission line) using 'Remove' button"""
        try:
            self.select_eline(eline)
            self.param_model.remove_peak_manual()
            self.apply_to_fit()
            self.fitting_parameters_changed()
        except Exception as ex:
            raise RuntimeError(str(ex))

    def update_userpeak(self, eline, energy, maxv, fwhm):
        """
        Update parameters of the user defined peak. Exception may be raised in
        case of an error. Note, that if energy is changed, FWHM will change to
        adjust for the variation of peak width.

        Parameters
        ----------
        eline: str
            name of the userpeak (e.g. 'Userpeak2')
        energy: float
            new position of the peak center, keV
        maxv: float
            new value of the amplitude
        fwhm: float
            new value of FWHM of the peak.
        """
        self.select_eline(eline)
        self.param_model.modify_userpeak_params(maxv, fwhm, energy)
        self.apply_to_fit()
        self.fitting_parameters_changed()

    def update_eline_peak_height(self, eline, maxv):
        self.select_eline(eline)
        self.param_model.modify_peak_height(maxv)
        self.apply_to_fit()
        self.fitting_parameters_changed()

    def get_suggested_manual_peak_energy(self):
        """
        Returns the suggested (pointed by the vertical marker) energy of the center peak in keV
        and the status of the marker.

        Returns
        -------
        float
            Energy of the manual peak center in keV. The energy is determined
            by vertical marker on the screen.
        bool
            True if the vertical marker is visible, otherwise False.

        """
        return self.plot_model.get_suggested_new_manual_peak_energy()

    def show_marker_at_current_position(self):
        """Force marker to visible state at the current position"""
        energy, marker_visible = self.get_suggested_manual_peak_energy()
        self.plot_model.set_plot_vertical_marker(energy)

    def get_peak_threshold(self):
        return self.param_model.bound_val

    def remove_peaks_below_threshold(self, peak_threshold):
        self.param_model.bound_val = peak_threshold
        self.param_model.remove_elements_below_threshold(threshv=peak_threshold)
        self.param_model.update_name_list()
        self.apply_to_fit()
        self.fitting_parameters_changed()

    def remove_unchecked_peaks(self):
        self.param_model.remove_elements_unselected()
        self.param_model.update_name_list()
        self.apply_to_fit()
        self.fitting_parameters_changed()

    def apply_to_fit(self):
        """
        Update plot, and apply updated parameters to fitting process.
        Note: this is an original function from 'fit.enaml' file.
        """
        self.param_model.EC.update_peak_ratio()
        self.param_model.data_for_plot()

        # update experimental plots in case the coefficients change
        self.plot_model.plot_experiment()

        self.plot_model.plot_fit(
            self.param_model.prefit_x, self.param_model.total_y, self.param_model.auto_fit_all
        )

        # For plotting purposes, otherwise plot will not update
        if self.plot_model.plot_exp_opt:
            self.plot_model.plot_exp_opt = False
            self.plot_model.plot_exp_opt = True
        else:
            self.plot_model.plot_exp_opt = True
            self.plot_model.plot_exp_opt = False
        self.plot_model.show_fit_opt = False
        self.plot_model.show_fit_opt = True

        # update parameter for fit
        self.param_model.create_full_param()  # Not sure this is needed
        self.fit_model.apply_default_param()

        # Update displayed intensity of the selected peak
        self.plot_model.compute_manual_peak_intensity()

    def total_spectrum_fitting(self):

        # Update parameter for fit. This may be unnecessary
        self.apply_to_fit()

        self.fit_model.fit_multiple()

        # BUG: line color for pileup is not correct from fit
        self.fit_model.get_profile()

        # update experimental plot with new calibration values
        self.plot_model.plot_experiment()

        self.plot_model.plot_fit(
            self.fit_model.cal_x, self.fit_model.cal_y, self.fit_model.cal_spectrum, self.fit_model.residual
        )

        # For plotting purposes, otherwise plot will not update
        self.plot_model.plot_exp_opt = False
        self.plot_model.plot_exp_opt = True
        self.plot_model.show_fit_opt = False
        self.plot_model.show_fit_opt = True

        # update autofit param
        self.param_model.update_new_param(self.param_model.param_new)
        # param_model.get_new_param_from_file(parameter_file_path)
        self.param_model.update_name_list()
        self.param_model.EC.turn_on_all()

        # Update displayed intensity of the selected peak
        self.plot_model.compute_manual_peak_intensity()

        self.fitting_parameters_changed()

    def save_param_to_file(self, path):
        save_as(path, self.param_model.param_new)

    def save_spectrum(self, dir, save_fit):
        self.fit_model.result_folder = dir
        self.fit_model.output_summed_data_fit(save_fit=save_fit)

    def compute_current_rfactor(self, save_fit):
        return self.fit_model.compute_current_rfactor(save_fit)

    def get_iter_and_var_number(self):
        return {"var_number": self.fit_model.nvar, "iter_number": self.fit_model.function_num}

    # ==========================================================================
    #          The following methods are used by Maps tab
    def fit_individual_pixels(self):
        self.apply_to_fit()

        self.fit_model.fit_single_pixel()

        # add scalers to fit dict
        scaler_keys = [v for v in self.io_model.img_dict.keys() if "scaler" in v]
        if len(scaler_keys) > 0:
            self.fit_model.fit_img[list(self.fit_model.fit_img.keys())[0]].update(
                self.io_model.img_dict[scaler_keys[0]]
            )

        self.io_model.update_img_dict(self.fit_model.fit_img)

    # ==========================================================================
    #          The following methods are used by ROI window
    def get_roi_selected_element_list(self):
        element_list = self.roi_model.element_for_roi
        # TODO: should probably be sorted differently
        return element_list

    def set_roi_selected_element_list(self, elements_for_roi):
        self.roi_model.element_for_roi = elements_for_roi
        self.plot_model.plot_roi_bound()

    def clear_roi_element_list(self):
        for r in self.roi_model.element_list_roi:
            self.plot_model.roi_dict[r].show_plot = False
        self.plot_model.plot_roi_bound()
        self.roi_model.clear_selected_elements()

    def load_roi_element_list_from_selected(self):
        for r in self.roi_model.element_list_roi:
            self.plot_model.roi_dict[r].show_plot = False
        self.plot_model.plot_roi_bound()
        selected_element_list = self.param_model.EC.get_element_list()
        selected_element_list = [
            _ for _ in selected_element_list if self.param_model.get_eline_name_category(_) == "eline"
        ]
        self.roi_model.select_elements_from_list(selected_element_list)

    def get_roi_subtract_background(self):
        return self.roi_model.subtract_background

    def set_roi_subtract_background(self, subtract_background):
        self.roi_model.subtract_background = bool(subtract_background)

    def compute_rois(self):
        if len(self.roi_model.element_for_roi) == 0:
            raise RuntimeError("No elements are selected for ROI computation. Select at least one element.")
        roi_result = self.roi_model.get_roi_sum()

        self.io_model.update_img_dict(roi_result)

    def show_roi(self, eline, show_status):
        self.plot_model.roi_dict = self.roi_model.roi_dict
        if show_status:
            # self.plot_model.plot_exp_opt = True
            self.plot_model.roi_dict[eline].show_plot = True
            self.plot_model.plot_roi_bound()
        else:
            self.plot_model.roi_dict[eline].show_plot = False
            self.plot_model.plot_roi_bound()

    def change_roi(self, eline, low, high):
        # Convert keV to eV (current implementation of ROIModel is using eV.
        self.roi_model.roi_dict[eline].left_val = int(low * 1000)
        self.roi_model.roi_dict[eline].right_val = int(high * 1000)
        self.plot_model.plot_roi_bound()

    def get_roi_settings(self):
        roi_settings = []

        eline_list = list(self.roi_model.element_list_roi)

        for eline in eline_list:
            # We display values in keV, but current implementation of ROIModel
            #   is using eV.
            energy_center = self.roi_model.roi_dict[eline].line_val / 1000.0
            energy_left = self.roi_model.roi_dict[eline].left_val / 1000.0
            energy_right = self.roi_model.roi_dict[eline].right_val / 1000.0
            range_displayed = self.roi_model.roi_dict[eline].show_plot

            energy_left_default = self.roi_model.roi_dict[eline].default_left / 1000.0
            energy_right_default = self.roi_model.roi_dict[eline].default_right / 1000.0

            roi_settings.append(
                {
                    "eline": eline,
                    "energy_center": energy_center,
                    "energy_left": energy_left,
                    "energy_right": energy_right,
                    "range_displayed": range_displayed,
                    "energy_left_default": energy_left_default,
                    "energy_right_default": energy_right_default,
                }
            )

        roi_settings.sort(key=lambda _: _["energy_center"])
        return roi_settings

    # ==========================================================================
    #          The following methods are used by ROI window
    def is_quant_standard_selected(self):
        """
        Returns True if quantitative standard was selected from the list.
        The result can be used to determine if saving quantitative standard is a valid
        operation.
        """
        return bool(self.fit_model.param_quant_estimation.fluorescence_data_dict)

    def is_quant_standard_fitting_available(self):
        """
        Returns True if fitted map is available.
        """
        result_map, selected_det_channel, scaler_name = self.fit_model.get_selected_fitted_map_data()
        return bool(result_map) and bool(selected_det_channel)

    def get_suggested_quant_file_name(self):
        """
        Return suggested 'standard' file name used for saving quantitative calibration
        to a JSON file
        """
        return self.fit_model.param_quant_estimation.get_suggested_json_fln()

    def get_displayed_quant_calib_parameters(self):

        fluorescence_data_dict = self.fit_model.param_quant_estimation.fluorescence_data_dict
        if fluorescence_data_dict is None:
            raise RuntimeError("Attempt to obtain quantitative standard parameters while no standard is selected")

        result_map, selected_det_channel, scaler_name = self.fit_model.get_selected_fitted_map_data()

        if not result_map:
            raise RuntimeError("No XRF maps were found. Run Individual Pixel Fitting to generate XRF maps.")

        if not selected_det_channel:
            raise RuntimeError("Selected XRF map does not contain fitted XRF maps.")

        suggested_fln = self.fit_model.param_quant_estimation.get_suggested_json_fln()

        # Fill the available fluorescence data
        self.fit_model.param_quant_estimation.fill_fluorescence_data_dict(
            xrf_map_dict=result_map, scaler_name=scaler_name
        )
        self.fit_model.param_quant_estimation.set_detector_channel_in_data_dict(
            detector_channel=selected_det_channel
        )

        distance_to_sample = self.fit_model.qe_standard_distance_to_sample
        self.fit_model.param_quant_estimation.set_distance_to_sample_in_data_dict(
            distance_to_sample=distance_to_sample
        )

        # Set optional parameters
        if self.io_model.scan_metadata_available:
            _scan_id = None
            _scan_uid = None
            if "scan_id" in self.io_model.scan_metadata:
                _scan_id = self.io_model.scan_metadata["scan_id"]
            if "scan_uid" in self.io_model.scan_metadata:
                _scan_uid = self.io_model.scan_metadata["scan_uid"]
            self.fit_model.param_quant_estimation.set_optional_parameters(scan_id=_scan_id, scan_uid=_scan_uid)

        preview = self.fit_model.param_quant_estimation.get_fluorescence_data_dict_text_preview()

        return {"suggested_file_name": suggested_fln, "distance_to_sample": distance_to_sample, "preview": preview}

    def save_distance_to_sample(self, distance_to_sample):
        self.fit_model.qe_standard_distance_to_sample = distance_to_sample
        self.fit_model.param_quant_estimation.set_distance_to_sample_in_data_dict(
            distance_to_sample=distance_to_sample
        )

    def save_quantitative_standard(self, file_path, overwrite_existing):
        """Will raise exception if unsuccessful"""
        self.fit_model.param_quant_estimation.save_fluorescence_data_dict(
            file_path, overwrite_existing=overwrite_existing
        )

    # ==========================================================================
    #     The following methods are used for exporting maps as TIFF and TXT
    def get_parameters_for_exporting_maps(self):
        dataset_list, dset_sel = self.get_maps_dataset_list()
        scalers, scaler_sel = self.get_maps_scaler_list()
        interpolate_on = self.get_maps_grid_interpolate()
        quant_norm_on = self.get_maps_quant_norm_enabled()

        return {
            "dset_list": dataset_list,
            "dset_sel": dset_sel,  # We want it to start from 0
            "scaler_list": scalers,
            "scaler_sel": scaler_sel,
            "interpolate_on": interpolate_on,
            "quant_norm_on": quant_norm_on,
        }

    def export_xrf_maps(
        self, *, results_path, dataset_name, scaler_name, interpolate_on, quant_norm_on, file_formats
    ):
        # We don't want to make any changes to 'param_quant_analysis' at the original location
        #   This is a temporary copy.
        param_quant_analysis = copy.deepcopy(self.img_model_adv.param_quant_analysis)
        param_quant_analysis.experiment_detector_channel = self.img_model_adv.get_detector_channel_name(
            dataset_name
        )
        for file_format in file_formats:
            self.fit_model.output_2Dimage(
                results_path=results_path,
                dataset_name=dataset_name,
                scaler_name=scaler_name,
                interpolate_on=interpolate_on,
                quant_norm_on=quant_norm_on,
                param_quant_analysis=param_quant_analysis,
                file_format=file_format,
            )

    # ==========================================================================
    #     Selection of the region pixels for saved spectra in Maps tab
    def get_enable_save_spectra(self):
        return self.fit_model.save_point

    def set_enable_save_spectra(self, enabled):
        self.fit_model.save_point = enabled

    def get_dataset_map_size(self):
        return self.io_model.get_dataset_map_size()

    def get_selection_area_save_spectra(self):
        return {
            "row_min": self.fit_model.point1v + 1,
            "row_max": self.fit_model.point2v,
            "col_min": self.fit_model.point1h + 1,
            "col_max": self.fit_model.point2h,
        }

    def set_selection_area_save_spectra(self, area):
        map_size = self.get_dataset_map_size()
        map_size = (1, 1) if map_size is None else map_size
        n_rows, n_cols = map_size

        def _fit_to_range(val, val_min, val_max):
            if val < val_min:
                val = val_min
            if val > val_max:
                val = val_max
            return val

        # Modify the selection so that at least one row or column is selected
        #   (area is at least 1 pixel wide in every direction)
        row_min = _fit_to_range(area["row_min"] - 1, 0, n_rows - 1)
        row_max = _fit_to_range(area["row_max"], row_min + 1, n_rows)
        col_min = _fit_to_range(area["col_min"] - 1, 0, n_cols - 1)
        col_max = _fit_to_range(area["col_max"], col_min + 1, n_cols)
        self.fit_model.point1v = row_min
        self.fit_model.point2v = row_max
        self.fit_model.point1h = col_min
        self.fit_model.point2h = col_max



from copy import deepcopy
from qtpy.QtWidgets import (
    QVBoxLayout,
    QHBoxLayout,
    QCheckBox,
    QPushButton,
    QHeaderView,
    QTableWidget,
    QTableWidgetItem,
    QSizePolicy,
)
from qtpy.QtGui import QBrush, QColor, QPalette
from qtpy.QtCore import Qt, Signal, Slot

from .useful_widgets import RangeManager, SecondaryWindow, set_tooltip

import logging

logger = logging.getLogger(__name__)


class WndImageWizard(SecondaryWindow):

    signal_redraw_maps = Signal()

    def __init__(self, *, gpc, gui_vars):
        super().__init__()

        # The variable enables/disables plot updates. Setting it False prevents
        #   plot updates while updating the table or Select/Deselect All operation
        self._enable_plot_updates = False
        self._changes_exist = False

        # Global processing classes
        self.gpc = gpc
        # Global GUI variables (used for control of GUI state)
        self.gui_vars = gui_vars

        self.initialize()

    def initialize(self):
        self.setWindowTitle("PyXRF: Image Wizard")

        self.setMinimumHeight(400)
        self.setMinimumWidth(500)
        self.resize(600, 600)

        self.cb_select_all = QCheckBox("All")
        self.cb_select_all.stateChanged.connect(self.cb_select_all_state_changed)

        self._auto_update = True
        self.cb_auto_update = QCheckBox("Auto")
        self.cb_auto_update.setCheckState(Qt.Checked if self._auto_update else Qt.Unchecked)
        self.cb_auto_update.stateChanged.connect(self.cb_auto_update_state_changed)

        self.pb_update_plots = QPushButton("Update Plots")
        self.pb_update_plots.setEnabled(not self._auto_update)
        self.pb_update_plots.clicked.connect(self.pb_update_plots_clicked)

        self._setup_table()

        vbox = QVBoxLayout()

        hbox = QHBoxLayout()
        hbox.addWidget(self.cb_select_all)
        hbox.addStretch(1)
        hbox.addWidget(self.cb_auto_update)
        hbox.addWidget(self.pb_update_plots)
        vbox.addLayout(hbox)

        hbox = QHBoxLayout()
        hbox.addWidget(self.table)
        vbox.addLayout(hbox)

        self.setLayout(vbox)

        self._set_tooltips()

    def _setup_table(self):

        self.table = QTableWidget()
        self._range_data = []  # The variable keeps copy of the table data
        self._limit_data = []  # Copy of the table that holds selection limits
        self._show_data = []  # Second column - bool values that indicate if the map is shown

        self.tbl_labels = ["Element", "Plotted Range", "Minimum", "Maximum"]
        self.tbl_h_alignment = [Qt.AlignCenter, Qt.AlignCenter, Qt.AlignCenter, Qt.AlignCenter]
        self.tbl_section_resize_mode = [
            QHeaderView.Stretch,
            QHeaderView.ResizeToContents,
            QHeaderView.Stretch,
            QHeaderView.Stretch,
        ]
        self.table.setColumnCount(len(self.tbl_labels))
        self.table.verticalHeader().hide()
        self.table.setHorizontalHeaderLabels(self.tbl_labels)

        self.table.setSelectionMode(QTableWidget.NoSelection)
        self.table.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)

        header = self.table.horizontalHeader()
        for n in range(len(self.tbl_labels)):
            header.setSectionResizeMode(n, self.tbl_section_resize_mode[n])
            header.setDefaultAlignment(self.tbl_h_alignment[n])

        self._checkable_items = []  # The list of items in the 1st column
        self._range_items = []  # The list of RangeManager items

        self.fill_table(self._range_data, self._limit_data, self._show_data)
        self.table.itemChanged.connect(self.table_item_changed)

    def fill_table(self, range_table, limit_table, show_table):

        self._enable_plot_updates = False

        self._clear_table()
        # Copy the table (we want to keep a copy, it's small)
        self._range_data = deepcopy(range_table)
        self._limit_data = deepcopy(limit_table)
        self._show_data = deepcopy(show_table)

        self.table.setRowCount(len(range_table))
        # Color is set for operation with Dark theme
        pal = self.table.palette()
        pal.setColor(QPalette.Text, Qt.black)
        pal.setColor(QPalette.Base, Qt.white)
        self.table.setPalette(pal)

        brightness = 200
        table_colors = [(255, brightness, brightness), (brightness, 255, brightness)]

        for nr, row in enumerate(range_table):
            element, v_min, v_max = row[0], row[1], row[2]
            sel_min, sel_max = self._limit_data[nr][1], self._limit_data[nr][2]
            sel_show = self._show_data[nr][1]
            rgb = table_colors[nr % 2]

            for nc in range(self.table.columnCount()):
                if nc in (0, 2, 3):

                    if nc == 0:
                        item = QTableWidgetItem(element)
                        item.setFlags(Qt.ItemIsUserCheckable | Qt.ItemIsEnabled)
                        item.setCheckState(Qt.Checked if sel_show else Qt.Unchecked)
                        self._checkable_items.append(item)
                    elif nc == 2:
                        item = QTableWidgetItem(f"{self._format_table_range_value(v_min)}")
                        item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
                    elif nc == 3:
                        item = QTableWidgetItem(f"{self._format_table_range_value(v_max)}")
                        item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)

                    item.setFlags(item.flags() & ~Qt.ItemIsEditable)

                    item.setBackground(QBrush(QColor(*rgb)))
                    self.table.setItem(nr, nc, item)

                elif nc == 1:

                    item = RangeManager(name=f"{nr}", add_sliders=True)
                    item.set_range(v_min, v_max)
                    item.set_selection(value_low=sel_min, value_high=sel_max)
                    item.setAlignment(Qt.AlignCenter)
                    item.setFixedWidth(400)
                    item.setBackground(rgb)
                    item.setTextColor([0, 0, 0])  # Color is set for operation with Dark theme
                    item.selection_changed.connect(self.table_range_item_changed)
                    self._range_items.append(item)
                    self.table.setCellWidget(nr, nc, item)

        self.table.resizeColumnsToContents()
        self.table.resizeRowsToContents()

        table_width = 0
        for n_col in range(self.table.columnCount()):
            table_width += self.table.columnWidth(n_col)
        self.table.setFixedWidth(table_width + 150)
        self.setFixedWidth(table_width + 170)

        self._enable_plot_updates = True

    def update_table_ranges(self, range_table, limit_table):
        """Update ranges and selections only. Don't update 'show' status."""
        self._enable_plot_updates = False

        self._range_data = deepcopy(range_table)
        self._limit_data = deepcopy(limit_table)

        for n in range(len(self._range_items)):
            v_min, v_max = self._range_data[n][1], self._range_data[n][2]
            sel_min, sel_max = self._limit_data[n][1], self._limit_data[n][2]
            rng = self._range_items[n]
            rng.set_range(v_min, v_max)
            rng.set_selection(value_low=sel_min, value_high=sel_max)

            self.table.item(n, 2).setText(f"{self._format_table_range_value(v_min)}")
            self.table.item(n, 3).setText(f"{self._format_table_range_value(v_max)}")

        self._enable_plot_updates = True

    def _set_tooltips(self):
        set_tooltip(self.cb_select_all, "<b>Select/Deselect All</b> emission lines in the list")
        set_tooltip(
            self.cb_auto_update,
            "Automatically <b>update the plots</b> when changes are made. "
            "If unchecked, then button <b>Update Plots</b> must be pressed "
            "to update the plots. Automatic update is often undesirable "
            "when large maps are displayed and multiple changes to parameters "
            "are made.",
        )
        set_tooltip(self.pb_update_plots, "<b>Update plots</b> based on currently selected parameters.")
        set_tooltip(
            self.table,
            "Choose <b>emission lines</b> from the currently selected dataset and "
            "select the <b>range of intensity</b> for each emission line",
        )

    def update_widget_state(self, condition=None):
        # Update the state of the menu bar
        state = not self.gui_vars["gui_state"]["running_computations"]
        self.setEnabled(state)

        # Hide the window if required by the program state
        state_xrf_map_exists = self.gui_vars["gui_state"]["state_xrf_map_exists"]
        if not state_xrf_map_exists:
            self.hide()

        if condition == "tooltips":
            self._set_tooltips()

    def cb_select_all_state_changed(self, state):
        self._select_all_items(state)

    def cb_auto_update_state_changed(self, state):
        self._auto_update = state
        self.pb_update_plots.setEnabled(not state)
        # If changes were made, apply the changes while switching to 'auto' mode
        if state and self._changes_exist:
            self._update_map_selections_auto()

    def pb_update_plots_clicked(self):
        """Upload the selections (limit table) and update plot"""
        self._update_map_selections()

    def table_item_changed(self, item):
        try:
            n_row = self._checkable_items.index(item)
            state = item.checkState() == Qt.Checked
            if state != self._show_data[n_row][1]:
                self._show_data[n_row][1] = state
                logger.debug(
                    f"Image wizard: map {self._show_data[n_row][0]} was {'checked' if state else 'unchecked'}"
                )
                self._update_map_selections_auto()
        except ValueError:
            pass

    def table_range_item_changed(self, low, high, name):
        n_row = int(name)
        self._limit_data[n_row][1] = low
        self._limit_data[n_row][2] = high
        logger.debug(
            f"Image Wizard: range changed for the map '{self._limit_data[n_row][0]}'. "
            f"New range: ({low}, {high})"
        )
        self._update_map_selections_auto()

    @Slot()
    def slot_update_table(self):
        """Reload table including ranges and selections for the emission lines"""
        range_table, limit_table, show_table = self.gpc.get_maps_info_table()
        self.fill_table(range_table, limit_table, show_table)

    @Slot()
    def slot_update_ranges(self):
        """Update only ranges and selections for the emission lines"""
        range_table, limit_table, _ = self.gpc.get_maps_info_table()
        self.update_table_ranges(range_table, limit_table)

    def _format_table_range_value(self, value):
        return f"{value:.12g}"

    def _clear_table(self):
        # Disconnect all signals
        for range_item in self._range_items:
            range_item.selection_changed.disconnect(self.table_range_item_changed)
        # Delete all rows
        self.table.clearContents()
        self._checkable_items = []  # The list of items in the 1st column
        self._range_items = []

    def _select_all_items(self, check_state):
        """Select/deselect all items in the table"""
        self._enable_plot_updates = False
        for n_row in range(self.table.rowCount()):
            item = self.table.item(n_row, 0)
            item.setCheckState(check_state)
        self._enable_plot_updates = True
        self._update_map_selections_auto()

    def _update_map_selections_auto(self):
        """Update maps only if 'auto' update is ON. Used as a 'filter'
        to prevent extra plot updates."""
        self._changes_exist = True
        if self._auto_update:
            self._update_map_selections()

    def _update_map_selections(self):
        """Upload the selections (limit table) and update plot"""
        if self._enable_plot_updates:
            self._changes_exist = False
            self.gpc.set_maps_limit_table(self._limit_data, self._show_data)
            self._redraw_maps()

    def _redraw_maps(self):
        logger.debug("Redrawing XRF Maps")
        self.gpc.redraw_maps()
        self.signal_redraw_maps.emit()



from pyxrf.gui_module.main_window import MainWindow
from PyQt5.QtWidgets import QMessageBox
from pyxrf.gui_support.gpc_class import GlobalProcessingClasses


def test_MainWindow(qtbot, monkeypatch):
    """
    Simple test that opens and closes the main window and
    checks the window dimensions
    """
    gpc = GlobalProcessingClasses()
    gpc.initialize()

    window = MainWindow(gpc=gpc)
    window.show()

    qtbot.addWidget(window)
    qtbot.waitForWindowShown(window)

    monkeypatch.setattr(QMessageBox, "exec", lambda *args: QMessageBox.No)
    window.close()
    assert not window._is_closed, "Window was closed prematurely"

    monkeypatch.setattr(QMessageBox, "exec", lambda *args: QMessageBox.Yes)
    window.close()
    assert window._is_closed, "Window was not closed properly"
